"""
This module extends asyncio.create_task to support any Awaitable,
manage task lifecycle, and enhance error handling.
"""

import asyncio
import logging

from a_sync import exceptions
from a_sync._typing import *


logger = logging.getLogger(__name__)


def create_task(
    coro: Awaitable[T],
    *,
    name: Optional[str] = None,
    skip_gc_until_done: bool = False,
    log_destroy_pending: bool = True,
) -> "asyncio.Task[T]":
    """
    Extends asyncio.create_task to support any Awaitable, manage task lifecycle, and enhance error handling.

    This function accepts any Awaitable, ensuring broader compatibility. If the Awaitable is not a coroutine,
    it attempts to convert it to one. It optionally prevents the task from being garbage-collected until completion
    and provides enhanced error management by wrapping exceptions in a custom exception when skip_gc_until_done is True.

    Args:
        coro: An Awaitable object from which to create the task. If not a coroutine, it will be converted.
        name: Optional name for the task, aiding in debugging.
        skip_gc_until_done: If True, the task is kept alive until it completes, preventing garbage collection.
                            Exceptions are wrapped in PersistedTaskException for special handling.
        log_destroy_pending: If False, asyncio's default error log when a pending task is destroyed is suppressed.

    Returns:
        An asyncio.Task object created from the provided Awaitable.
    """

    if not asyncio.iscoroutine(coro):
        coro = __await(coro)
    task = asyncio.create_task(coro, name=name)
    if skip_gc_until_done:
        __persisted_tasks.add(asyncio.create_task(__persisted_task_exc_wrap(task)))
    if log_destroy_pending is False:
        asyncio.Task.__del__
        task._log_destroy_pending = False
    __prune_persisted_tasks()
    return task


__persisted_tasks: Set["asyncio.Task[Any]"] = set()


async def __await(awaitable: Awaitable[T]) -> T:
    """Wait for the completion of an Awaitable.

    Args:
        awaitable: The Awaitable object to wait for.

    Raises:
        RuntimeError: If a RuntimeError occurs during the await, it is raised with additional context.
    """
    try:
        return await awaitable
    except RuntimeError as e:
        args = [e, awaitable]
        if isinstance(awaitable, asyncio.tasks._GatheringFuture):
            args.append(awaitable._children)
        raise RuntimeError(*args) from None


def __prune_persisted_tasks():
    """Remove completed tasks from the set of persisted tasks.

    This function checks each task in the persisted tasks set. If a task is done and has an exception,
    it logs the exception and raises it if it's not a PersistedTaskException. It also logs the traceback
    manually since the usual handler will not run after retrieving the exception.
    """
    for task in tuple(__persisted_tasks):
        if task.done() and (e := task.exception()):
            # force exceptions related to this lib to bubble up
            if not isinstance(e, exceptions.PersistedTaskException):
                logger.exception(e)
                raise e
            # we have to manually log the traceback that asyncio would usually log
            # since we already got the exception from the task and the usual handler will now not run
            context = {
                "message": f"{task.__class__.__name__} exception was never retrieved",
                "exception": e,
                "future": task,
            }
            if task._source_traceback:
                context["source_traceback"] = task._source_traceback
            task._loop.call_exception_handler(context)
            __persisted_tasks.discard(task)


async def __persisted_task_exc_wrap(task: "asyncio.Task[T]") -> T:
    """
    Wrap a task to handle its exception in a specialized manner.

    Args:
        task: The asyncio Task to wrap.

    Raises:
        PersistedTaskException: Wraps any exception raised by the task for special handling.
    """
    try:
        return await task
    except Exception as e:
        raise exceptions.PersistedTaskException(e, task) from e


__all__ = ["create_task"]



"""
This module defines custom exceptions for the a_sync library.
"""

import asyncio

from a_sync._typing import *
from a_sync.a_sync._flags import VIABLE_FLAGS

if TYPE_CHECKING:
    from a_sync import TaskMapping


class ASyncFlagException(ValueError):
    """
    Base exception class for flag-related errors in the a_sync library.
    """

    viable_flags = VIABLE_FLAGS
    """
    The set of viable flags.

    A-Sync uses 'flags' to indicate whether objects / fn calls will be sync or async.
    You can use any of the provided flags, whichever makes most sense for your use case.
    """

    def desc(self, target) -> str:
        """
        Returns a description of the target for the flag error message.

        Args:
            target: The target object or string to describe.

        Returns:
            A string description of the target.
        """
        if target == "kwargs":
            return "flags present in 'kwargs'"
        else:
            return f"flag attributes defined on {target}"


class NoFlagsFound(ASyncFlagException):
    """
    Raised when no viable flags are found in the target.
    """

    def __init__(self, target, kwargs_keys=None):
        """
        Initializes the NoFlagsFound exception.

        Args:
            target: The target object where flags were expected.
            kwargs_keys: Optional; keys in the kwargs if applicable.
        """
        err = f"There are no viable a_sync {self.desc(target)}:"
        err += f"\nViable flags: {self.viable_flags}"
        if kwargs_keys:
            err += f"\nkwargs keys: {kwargs_keys}"
        err += "\nThis is likely an issue with a custom subclass definition."
        super().__init__(err)


class TooManyFlags(ASyncFlagException):
    """
    Raised when multiple flags are found, but only one was expected.
    """

    def __init__(self, target, present_flags):
        """
        Initializes the TooManyFlags exception.

        Args:
            target: The target object where flags were found.
            present_flags: The flags that were found.
        """
        err = f"There are multiple a_sync {self.__get_desc(target)} and there should only be one.\n"
        err += f"Present flags: {present_flags}\n"
        err += "This is likely an issue with a custom subclass definition."
        super().__init__(err)


class InvalidFlag(ASyncFlagException):
    """
    Raised when an invalid flag is encountered.
    """

    def __init__(self, flag: Optional[str]):
        """
        Initializes the InvalidFlag exception.

        Args:
            flag: The invalid flag.
        """
        err = f"'flag' must be one of: {self.viable_flags}. You passed {flag}."
        err += "\nThis code should not be reached and likely indicates an issue with a custom subclass definition."
        super().__init__(err)


class InvalidFlagValue(ASyncFlagException):
    """
    Raised when a flag has an invalid value.
    """

    def __init__(self, flag: str, flag_value: Any):
        """
        Initializes the InvalidFlagValue exception.

        Args:
            flag: The flag with an invalid value.
            flag_value: The invalid value of the flag.
        """
        super().__init__(f"'{flag}' should be boolean. You passed {flag_value}.")


class FlagNotDefined(ASyncFlagException):
    """
    Raised when a flag is not defined on an object.
    """

    def __init__(self, obj: Type, flag: str):
        """
        Initializes the FlagNotDefined exception.

        Args:
            obj: The object where the flag is not defined.
            flag: The undefined flag.
        """
        super().__init__(f"{obj} flag {flag} is not defined.")


class ImproperFunctionType(ValueError):
    """
    Raised when a function that should be sync is async or vice-versa.
    """


class FunctionNotAsync(ImproperFunctionType):
    """
    Raised when a function expected to be async is not.
    """

    def __init__(self, fn):
        """
        Initializes the FunctionNotAsync exception.

        Args:
            fn: The function that is not async.
        """
        super().__init__(
            f"`coro_fn` must be a coroutine function defined with `async def`. You passed {fn}."
        )


class FunctionNotSync(ImproperFunctionType):
    """
    Raised when a function expected to be sync is not.
    """

    def __init__(self, fn):
        """
        Initializes the FunctionNotSync exception.

        Args:
            fn: The function that is not sync.
        """
        super().__init__(
            f"`func` must be a coroutine function defined with `def`. You passed {fn}."
        )


class ASyncRuntimeError(RuntimeError):
    """
    Raised for runtime errors in asynchronous operations.
    """

    def __init__(self, e: RuntimeError):
        """
        Initializes the ASyncRuntimeError exception.

        Args:
            e: The original runtime error.
        """
        super().__init__(str(e))


class SyncModeInAsyncContextError(ASyncRuntimeError):
    """
    Raised when synchronous code is used within an asynchronous context.
    """

    def __init__(self, err: str = ""):
        """
        Initializes the SyncModeInAsyncContextError exception.
        """
        if not err:
            err = "The event loop is already running, which means you're trying to use an `ASyncFunction` synchronously from within an async context.\n"
            err += f"Check your traceback to determine which, then try calling asynchronously instead with one of the following kwargs:\n"
            err += f"{VIABLE_FLAGS}"
        super().__init__(err)


class MappingError(Exception):
    """
    Base class for errors related to :class:`~TaskMapping`.
    """

    _msg: str

    def __init__(self, mapping: "TaskMapping", msg: str = ""):
        """
        Initializes the MappingError exception.

        Args:
            mapping: The TaskMapping where the error occurred.
            msg: An optional message describing the error.
        """
        msg = msg or self._msg + f":\n{mapping}"
        if mapping:
            msg += f"\n{dict(mapping)}"
        super().__init__(msg)
        self.mapping = mapping


class MappingIsEmptyError(MappingError):
    """
    Raised when a TaskMapping is empty and an operation requires it to have items.
    """

    _msg = "TaskMapping does not contain anything to yield"


class MappingNotEmptyError(MappingError):
    """
    Raised when a TaskMapping is not empty and an operation requires it to be empty.
    """

    _msg = "TaskMapping already contains some data. In order to use `map`, you need a fresh one"


class PersistedTaskException(Exception):
    """
    Raised when an exception persists in an asyncio Task.
    """

    def __init__(self, exc: E, task: asyncio.Task) -> None:
        """
        Initializes the PersistedTaskException exception.

        Args:
            exc: The exception that persisted.
            task: The asyncio Task where the exception occurred.
        """
        super().__init__(f"{exc.__class__.__name__}: {exc}", task)
        self.exception = exc
        self.task = task


class EmptySequenceError(ValueError):
    """
    Raised when an operation is attempted on an empty sequence but items are required.
    """



# mypy: disable-error-code=valid-type
# mypy: disable-error-code=misc
import asyncio

from a_sync import exceptions
from a_sync._typing import *
from a_sync.a_sync.modifiers.cache.memory import apply_async_memory_cache


class CacheArgs(TypedDict):
    """Typed dictionary for cache arguments."""

    cache_type: CacheType
    cache_typed: bool
    ram_cache_maxsize: Optional[int]
    ram_cache_ttl: Optional[int]


@overload
def apply_async_cache(
    **modifiers: Unpack[CacheArgs],
) -> AsyncDecorator[P, T]:
    """Overload for when no coroutine function is provided."""


@overload
def apply_async_cache(
    coro_fn: int,
    **modifiers: Unpack[CacheArgs],
) -> AsyncDecorator[P, T]:
    """Overload for when an integer is provided as the coroutine function."""


@overload
def apply_async_cache(
    coro_fn: CoroFn[P, T],
    **modifiers: Unpack[CacheArgs],
) -> CoroFn[P, T]:
    """Overload for when a coroutine function is provided."""


def apply_async_cache(
    coro_fn: Union[CoroFn[P, T], CacheType, int] = None,
    cache_type: CacheType = "memory",
    cache_typed: bool = False,
    ram_cache_maxsize: Optional[int] = None,
    ram_cache_ttl: Optional[int] = None,
) -> AsyncDecoratorOrCoroFn[P, T]:
    """Applies an asynchronous cache to a coroutine function.

    Args:
        coro_fn: The coroutine function to apply the cache to, or an integer to set as the max size.
        cache_type: The type of cache to use. Currently, only 'memory' is implemented.
        cache_typed: Whether to consider types for cache keys.
        ram_cache_maxsize: The maximum size for the LRU cache. If set to an integer, it overrides coro_fn.
        ram_cache_ttl: The time-to-live for items in the LRU cache.

    Raises:
        TypeError: If 'ram_cache_maxsize' is not an integer or None.
        FunctionNotAsync: If the provided function is not asynchronous.
        NotImplementedError: If an unsupported cache type is specified.

    Returns:
        A decorator or the decorated coroutine function.
    """

    # Parse Inputs
    if isinstance(coro_fn, int):
        assert ram_cache_maxsize is None
        ram_cache_maxsize = coro_fn
        coro_fn = None

    # Validate
    elif coro_fn is None:
        if ram_cache_maxsize is not None and not isinstance(ram_cache_maxsize, int):
            raise TypeError(
                "'lru_cache_maxsize' must be an integer or None.", ram_cache_maxsize
            )
    elif not asyncio.iscoroutinefunction(coro_fn):
        raise exceptions.FunctionNotAsync(coro_fn)

    if cache_type == "memory":
        cache_decorator = apply_async_memory_cache(
            maxsize=ram_cache_maxsize, ttl=ram_cache_ttl, typed=cache_typed
        )
        return cache_decorator if coro_fn is None else cache_decorator(coro_fn)
    elif cache_type == "disk":
        pass
    raise NotImplementedError(f"cache_type: {cache_type}")



import asyncio


def get_event_loop() -> asyncio.AbstractEventLoop:
    try:
        loop = asyncio.get_event_loop()
    except RuntimeError as e:  # Necessary for use with multi-threaded applications.
        if not str(e).startswith("There is no current event loop in thread"):
            raise
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
    return loop



"""
This module provides utility functions for handling keyword arguments related to synchronous and asynchronous flags.
"""

from typing import Optional

from a_sync import exceptions
from a_sync.a_sync import _flags


def get_flag_name(kwargs: dict) -> Optional[str]:
    """
    Get the name of the flag present in the kwargs.

    Args:
        kwargs: A dictionary of keyword arguments.

    Returns:
        The name of the flag if present, None otherwise.

    Raises:
        :class:`exceptions.TooManyFlags`: If more than one flag is present in the kwargs.
    """
    present_flags = [flag for flag in _flags.VIABLE_FLAGS if flag in kwargs]
    if len(present_flags) == 0:
        return None
    if len(present_flags) != 1:
        raise exceptions.TooManyFlags("kwargs", present_flags)
    return present_flags[0]


def is_sync(flag: str, kwargs: dict, pop_flag: bool = False) -> bool:
    """
    Determine if the operation should be synchronous based on the flag value.

    Args:
        flag: The name of the flag to check.
        kwargs: A dictionary of keyword arguments.
        pop_flag: Whether to remove the flag from kwargs. Defaults to False.

    Returns:
        True if the operation should be synchronous, False otherwise.
    """
    flag_value = kwargs.pop(flag) if pop_flag else kwargs[flag]
    return _flags.negate_if_necessary(flag, flag_value)



import asyncio
import time

import pytest

from a_sync import ProcessPoolExecutor, ThreadPoolExecutor, PruningThreadPoolExecutor


@pytest.mark.asyncio
async def test_process_pool_executor_run():
    """Tests the ProcessPoolExecutor by running and submitting the work function asynchronously."""
    executor = ProcessPoolExecutor(1)
    coro = executor.run(time.sleep, 0.1)
    assert asyncio.iscoroutine(coro)
    await coro


@pytest.mark.asyncio
async def test_thread_pool_executor_run():
    """Tests the ThreadPoolExecutor by running and submitting the work function asynchronously."""
    executor = ThreadPoolExecutor(1)
    coro = executor.run(time.sleep, 0.1)
    assert asyncio.iscoroutine(coro)
    await coro


@pytest.mark.asyncio
async def test_pruning_thread_pool_executor_run():
    """Tests the PruningThreadPoolExecutor by running and submitting the work function asynchronously."""
    executor = PruningThreadPoolExecutor(1)
    coro = executor.run(time.sleep, 0.1)
    assert asyncio.iscoroutine(coro)
    await coro


@pytest.mark.asyncio
async def test_process_pool_executor_submit():
    """Tests the ProcessPoolExecutor by submitting the work function asynchronously."""
    executor = ProcessPoolExecutor(1)
    fut = executor.submit(time.sleep, 0.1)
    assert isinstance(fut, asyncio.Future)
    await fut


@pytest.mark.asyncio
async def test_thread_pool_executor_submit():
    """Tests the ThreadPoolExecutor by submitting the work function asynchronously."""
    executor = ThreadPoolExecutor(1)
    fut = executor.submit(time.sleep, 0.1)
    assert isinstance(fut, asyncio.Future)
    await fut


@pytest.mark.asyncio
async def test_pruning_thread_pool_executor_submit():
    """Tests the PruningThreadPoolExecutor by submitting the work function asynchronously."""
    executor = PruningThreadPoolExecutor(1)
    fut = executor.submit(time.sleep, 0.1)
    assert isinstance(fut, asyncio.Future)
    await fut


@pytest.mark.asyncio
async def test_process_pool_executor_sync_run():
    """Tests the ProcessPoolExecutor by running and submitting the work function synchronously."""
    executor = ProcessPoolExecutor(0)
    coro = executor.run(time.sleep, 0.1)
    assert asyncio.iscoroutine(coro)
    await coro


@pytest.mark.asyncio
async def test_thread_pool_executor_sync_run():
    """Tests the ThreadPoolExecutor by running and submitting the work function synchronously."""
    executor = ThreadPoolExecutor(0)
    coro = executor.run(time.sleep, 0.1)
    assert asyncio.iscoroutine(coro)
    await coro


@pytest.mark.asyncio
async def test_pruning_thread_pool_executor_sync_run():
    """Tests the PruningThreadPoolExecutor by running and submitting the work function synchronously."""
    executor = PruningThreadPoolExecutor(0)
    coro = executor.run(time.sleep, 0.1)
    assert asyncio.iscoroutine(coro)
    await coro


@pytest.mark.asyncio
async def test_process_pool_executor_sync_submit():
    """Tests the ProcessPoolExecutor by submitting the work function synchronously."""
    executor = ProcessPoolExecutor(0)
    fut = executor.submit(time.sleep, 0.1)
    assert isinstance(fut, asyncio.Future)
    await fut


@pytest.mark.asyncio
async def test_thread_pool_executor_sync_submit():
    """Tests the ThreadPoolExecutor by submitting the work function synchronously."""
    executor = ThreadPoolExecutor(0)
    fut = executor.submit(time.sleep, 0.1)
    assert isinstance(fut, asyncio.Future)
    await fut


@pytest.mark.asyncio
async def test_pruning_thread_pool_executor_sync_submit():
    """Tests the PruningThreadPoolExecutor by submitting the work function synchronously."""
    executor = PruningThreadPoolExecutor(0)
    fut = executor.submit(time.sleep, 0.1)
    assert isinstance(fut, asyncio.Future)
    await fut



"""
This module provides an enhanced version of asyncio.Event with additional debug logging to help detect deadlocks.
"""

import asyncio
import sys

from a_sync._typing import *
from a_sync.primitives._debug import _DebugDaemonMixin


class Event(asyncio.Event, _DebugDaemonMixin):
    """
    An asyncio.Event with additional debug logging to help detect deadlocks.

    This event class extends asyncio.Event by adding debug logging capabilities. It logs
    detailed information about the event state and waiters, which can be useful for
    diagnosing and debugging potential deadlocks.
    """

    _value: bool
    _loop: asyncio.AbstractEventLoop
    _waiters: Deque["asyncio.Future[None]"]
    if sys.version_info >= (3, 10):
        __slots__ = "_value", "_waiters", "_debug_daemon_interval"
    else:
        __slots__ = "_value", "_loop", "_waiters", "_debug_daemon_interval"

    def __init__(
        self,
        name: str = "",
        debug_daemon_interval: int = 300,
        *,
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ):
        """
        Initializes the Event.

        Args:
            name (str): An optional name for the event, used in debug logs.
            debug_daemon_interval (int): The interval in seconds for the debug daemon to log information.
            loop (Optional[asyncio.AbstractEventLoop]): The event loop to use.
        """
        if sys.version_info >= (3, 10):
            super().__init__()
        else:
            super().__init__(loop=loop)
        self._name = name
        # backwards compatability
        if hasattr(self, "_loop"):
            self._loop = self._loop or asyncio.get_event_loop()
        self._debug_daemon_interval = debug_daemon_interval

    def __repr__(self) -> str:
        label = f"name={self._name}" if self._name else "object"
        status = "set" if self._value else "unset"
        if self._waiters:
            status += f", waiters:{len(self._waiters)}"
        return f"<{self.__class__.__module__}.{self.__class__.__name__} {label} at {hex(id(self))} [{status}]>"

    async def wait(self) -> Literal[True]:
        """
        Wait until the event is set.

        Returns:
            True when the event is set.
        """
        if self.is_set():
            return True
        self._ensure_debug_daemon()
        return await super().wait()

    async def _debug_daemon(self) -> None:
        """
        Periodically logs debug information about the event state and waiters.
        """
        weakself = weakref.ref(self)
        del self  # no need to hold a reference here
        while (self := weakself()) and not self.is_set():
            del self  # no need to hold a reference here
            await asyncio.sleep(self._debug_daemon_interval)
            if (self := weakself()) and not self.is_set():
                self.logger.debug(
                    "Waiting for %s for %sm", self, round((time() - start) / 60, 2)
                )



# type: ignore [var-annotated]

"""
The future.py module provides functionality for handling asynchronous futures, 
including a decorator for converting callables into ASyncFuture objects and 
utilities for managing asynchronous computations.

Functions:
    future(callable: Union[Callable[P, Awaitable[T]], Callable[P, T]] = None, **kwargs: Unpack[ModifierKwargs]) -> Callable[P, Union[T, "ASyncFuture[T]"]]: 
        A decorator to convert a callable into an ASyncFuture, with optional modifiers.
    _gather_check_and_materialize(*things: Unpack[MaybeAwaitable[T]]) -> List[T]: 
        Gathers and materializes a list of awaitable or non-awaitable items.
    _check_and_materialize(thing: T) -> T: 
        Checks if an item is awaitable and materializes it.
    _materialize(meta: "ASyncFuture[T]") -> T: 
        Materializes the result of an ASyncFuture.

Classes:
    ASyncFuture: Represents an asynchronous future result.
    _ASyncFutureWrappedFn: A callable class to wrap functions and return ASyncFuture objects.
    _ASyncFutureInstanceMethod: A class to handle instance methods wrapped as ASyncFuture.
"""

import asyncio
import concurrent.futures
from functools import partial, wraps
from inspect import isawaitable

from a_sync._typing import *


def future(
    callable: AnyFn[P, T] = None,
    **kwargs: Unpack[ModifierKwargs],
) -> Callable[P, Union[T, "ASyncFuture[T]"]]:
    """
    A decorator function to convert a callable into an ASyncFuture, with optional modifiers.

    Args:
        callable: The callable to convert. Defaults to None.
        **kwargs: Additional keyword arguments for the modifier.
    """
    return _ASyncFutureWrappedFn(callable, **kwargs)


async def _gather_check_and_materialize(*things: Unpack[MaybeAwaitable[T]]) -> List[T]:
    """
    Gathers and materializes a list of awaitable or non-awaitable items.

    Args:
        *things: Items to gather and materialize.
    """
    return await asyncio.gather(*[_check_and_materialize(thing) for thing in things])


async def _check_and_materialize(thing: T) -> T:
    """
    Checks if an item is awaitable and materializes it.

    Args:
        thing: The item to check and materialize.
    """
    return await thing if isawaitable(thing) else thing


def _materialize(meta: "ASyncFuture[T]") -> T:
    """
    Materializes the result of an ASyncFuture.

    Args:
        meta: The ASyncFuture to materialize.

    Raises:
        RuntimeError: If the event loop is running and the result cannot be awaited.
    """
    try:
        return asyncio.get_event_loop().run_until_complete(meta)
    except RuntimeError as e:
        raise RuntimeError(
            f"{meta} result is not set and the event loop is running, you will need to await it first"
        ) from e


MetaNumeric = Union[
    Numeric, "ASyncFuture[int]", "ASyncFuture[float]", "ASyncFuture[Decimal]"
]


class ASyncFuture(concurrent.futures.Future, Awaitable[T]):
    """
    A class representing an asynchronous future result.

    Inherits from both concurrent.futures.Future and Awaitable[T], allowing it to be used in both synchronous and asynchronous contexts.
    """

    __slots__ = "__awaitable__", "__dependencies", "__dependants", "__task"

    def __init__(
        self, awaitable: Awaitable[T], dependencies: List["ASyncFuture"] = []
    ) -> None:
        """
        Initializes an ASyncFuture with an awaitable and optional dependencies.

        Args:
            awaitable: The awaitable object.
            dependencies: A list of dependencies. Defaults to [].
        """
        self.__awaitable__ = awaitable
        """The awaitable object."""
        self.__dependencies = dependencies
        """A list of dependencies."""
        for dependency in dependencies:
            assert isinstance(dependency, ASyncFuture)
            dependency.__dependants.append(self)
        self.__dependants: List[ASyncFuture] = []
        """A list of dependants."""
        self.__task = None
        """The task associated with the awaitable."""
        super().__init__()

    def __hash__(self) -> int:
        return hash(self.__awaitable__)

    def __repr__(self) -> str:
        string = f"<{self.__class__.__name__} {self._state} for {self.__awaitable__}"
        if self.cancelled():
            pass
        elif self.done():
            string += (
                f" exception={self.exception()}"
                if self.exception()
                else f" result={super().result()}"
            )
        return string + ">"

    def __list_dependencies(self, other) -> List["ASyncFuture"]:
        """
        Lists dependencies for the ASyncFuture.

        Args:
            other: The other dependency to list.
        """
        if isinstance(other, ASyncFuture):
            return [self, other]
        return [self]

    @property
    def result(self) -> Union[Callable[[], T], Any]:
        """
        If this future is not done, it will work like cf.Future.result. It will block, await the awaitable, and return the result when ready.
        If this future is done and the result has attribute `result`, will return `getattr(future_result, 'result')`
        If this future is done and the result does NOT have attribute `result`, will again work like cf.Future.result
        """
        if self.done():
            if hasattr(r := super().result(), "result"):
                # can be property, method, whatever. should work.
                return r.result
            # the result should be callable like an asyncio.Future
            return super().result
        return lambda: _materialize(self)

    def __getattr__(self, attr: str) -> Any:
        return getattr(_materialize(self), attr)

    def __getitem__(self, key) -> Any:
        return _materialize(self)[key]

    # NOTE: broken, do not use. I think
    def __setitem__(self, key, value) -> None:
        _materialize(self)[key] = value

    # not sure what to call these
    def __contains__(self, key: Any) -> bool:
        return _materialize(
            ASyncFuture(
                self.__contains(key), dependencies=self.__list_dependencies(key)
            )
        )

    def __await__(self) -> Generator[Any, None, T]:
        """
        Makes the ASyncFuture awaitable.
        """
        return self.__await().__await__()

    async def __await(self) -> T:
        if not self.done():
            self.set_result(await self.__task__)
        return self._result

    @property
    def __task__(self) -> "asyncio.Task[T]":
        """
        Returns the asyncio task associated with the awaitable, creating it if necessary.
        """
        if self.__task is None:
            self.__task = asyncio.create_task(self.__awaitable__)
        return self.__task

    def __iter__(self):
        return _materialize(self).__iter__()

    def __next__(self):
        return _materialize(self).__next__()

    def __enter__(self):
        return _materialize(self).__enter__()

    def __exit__(self, *args):
        return _materialize(self).__exit__(*args)

    @overload
    def __add__(self: "ASyncFuture[int]", other: int) -> "ASyncFuture[int]": ...
    @overload
    def __add__(self: "ASyncFuture[float]", other: float) -> "ASyncFuture[float]": ...
    @overload
    def __add__(self: "ASyncFuture[float]", other: int) -> "ASyncFuture[float]": ...
    @overload
    def __add__(self: "ASyncFuture[int]", other: float) -> "ASyncFuture[float]": ...
    @overload
    def __add__(
        self: "ASyncFuture[Decimal]", other: Decimal
    ) -> "ASyncFuture[Decimal]": ...
    @overload
    def __add__(self: "ASyncFuture[Decimal]", other: int) -> "ASyncFuture[Decimal]": ...
    @overload
    def __add__(self: "ASyncFuture[int]", other: Decimal) -> "ASyncFuture[Decimal]": ...
    @overload
    def __add__(
        self: "ASyncFuture[int]", other: Awaitable[int]
    ) -> "ASyncFuture[int]": ...
    @overload
    def __add__(
        self: "ASyncFuture[float]", other: Awaitable[float]
    ) -> "ASyncFuture[float]": ...
    @overload
    def __add__(
        self: "ASyncFuture[float]", other: Awaitable[int]
    ) -> "ASyncFuture[float]": ...
    @overload
    def __add__(
        self: "ASyncFuture[int]", other: Awaitable[float]
    ) -> "ASyncFuture[float]": ...
    @overload
    def __add__(
        self: "ASyncFuture[Decimal]", other: Awaitable[Decimal]
    ) -> "ASyncFuture[Decimal]": ...
    @overload
    def __add__(
        self: "ASyncFuture[Decimal]", other: Awaitable[int]
    ) -> "ASyncFuture[Decimal]": ...
    @overload
    def __add__(
        self: "ASyncFuture[int]", other: Awaitable[Decimal]
    ) -> "ASyncFuture[Decimal]": ...
    def __add__(self, other: MetaNumeric) -> "ASyncFuture":
        return ASyncFuture(
            self.__add(other), dependencies=self.__list_dependencies(other)
        )

    @overload
    def __sub__(self: "ASyncFuture[int]", other: int) -> "ASyncFuture[int]": ...
    @overload
    def __sub__(self: "ASyncFuture[float]", other: float) -> "ASyncFuture[float]": ...
    @overload
    def __sub__(self: "ASyncFuture[float]", other: int) -> "ASyncFuture[float]": ...
    @overload
    def __sub__(self: "ASyncFuture[int]", other: float) -> "ASyncFuture[float]": ...
    @overload
    def __sub__(
        self: "ASyncFuture[Decimal]", other: Decimal
    ) -> "ASyncFuture[Decimal]": ...
    @overload
    def __sub__(self: "ASyncFuture[Decimal]", other: int) -> "ASyncFuture[Decimal]": ...
    @overload
    def __sub__(self: "ASyncFuture[int]", other: Decimal) -> "ASyncFuture[Decimal]": ...
    @overload
    def __sub__(
        self: "ASyncFuture[int]", other: Awaitable[int]
    ) -> "ASyncFuture[int]": ...
    @overload
    def __sub__(
        self: "ASyncFuture[float]", other: Awaitable[float]
    ) -> "ASyncFuture[float]": ...
    @overload
    def __sub__(
        self: "ASyncFuture[float]", other: Awaitable[int]
    ) -> "ASyncFuture[float]": ...
    @overload
    def __sub__(
        self: "ASyncFuture[int]", other: Awaitable[float]
    ) -> "ASyncFuture[float]": ...
    @overload
    def __sub__(
        self: "ASyncFuture[Decimal]", other: Awaitable[Decimal]
    ) -> "ASyncFuture[Decimal]": ...
    @overload
    def __sub__(
        self: "ASyncFuture[Decimal]", other: Awaitable[int]
    ) -> "ASyncFuture[Decimal]": ...
    @overload
    def __sub__(
        self: "ASyncFuture[int]", other: Awaitable[Decimal]
    ) -> "ASyncFuture[Decimal]": ...
    def __sub__(self, other: MetaNumeric) -> "ASyncFuture":
        return ASyncFuture(
            self.__sub(other), dependencies=self.__list_dependencies(other)
        )

    def __mul__(self, other) -> "ASyncFuture":
        return ASyncFuture(
            self.__mul(other), dependencies=self.__list_dependencies(other)
        )

    def __pow__(self, other) -> "ASyncFuture":
        return ASyncFuture(
            self.__pow(other), dependencies=self.__list_dependencies(other)
        )

    def __truediv__(self, other) -> "ASyncFuture":
        return ASyncFuture(
            self.__truediv(other), dependencies=self.__list_dependencies(other)
        )

    def __floordiv__(self, other) -> "ASyncFuture":
        return ASyncFuture(
            self.__floordiv(other), dependencies=self.__list_dependencies(other)
        )

    def __pow__(self, other) -> "ASyncFuture":
        return ASyncFuture(
            self.__pow(other), dependencies=self.__list_dependencies(other)
        )

    @overload
    def __radd__(self: "ASyncFuture[int]", other: int) -> "ASyncFuture[int]": ...
    @overload
    def __radd__(self: "ASyncFuture[float]", other: float) -> "ASyncFuture[float]": ...
    @overload
    def __radd__(self: "ASyncFuture[float]", other: int) -> "ASyncFuture[float]": ...
    @overload
    def __radd__(self: "ASyncFuture[int]", other: float) -> "ASyncFuture[float]": ...
    @overload
    def __radd__(
        self: "ASyncFuture[Decimal]", other: Decimal
    ) -> "ASyncFuture[Decimal]": ...
    @overload
    def __radd__(
        self: "ASyncFuture[Decimal]", other: int
    ) -> "ASyncFuture[Decimal]": ...
    @overload
    def __radd__(
        self: "ASyncFuture[int]", other: Decimal
    ) -> "ASyncFuture[Decimal]": ...
    @overload
    def __radd__(
        self: "ASyncFuture[int]", other: Awaitable[int]
    ) -> "ASyncFuture[int]": ...
    @overload
    def __radd__(
        self: "ASyncFuture[float]", other: Awaitable[float]
    ) -> "ASyncFuture[float]": ...
    @overload
    def __radd__(
        self: "ASyncFuture[float]", other: Awaitable[int]
    ) -> "ASyncFuture[float]": ...
    @overload
    def __radd__(
        self: "ASyncFuture[int]", other: Awaitable[float]
    ) -> "ASyncFuture[float]": ...
    @overload
    def __radd__(
        self: "ASyncFuture[Decimal]", other: Awaitable[Decimal]
    ) -> "ASyncFuture[Decimal]": ...
    @overload
    def __radd__(
        self: "ASyncFuture[Decimal]", other: Awaitable[int]
    ) -> "ASyncFuture[Decimal]": ...
    @overload
    def __radd__(
        self: "ASyncFuture[int]", other: Awaitable[Decimal]
    ) -> "ASyncFuture[Decimal]": ...
    def __radd__(self, other) -> "ASyncFuture":
        return ASyncFuture(
            self.__radd(other), dependencies=self.__list_dependencies(other)
        )

    @overload
    def __rsub__(self: "ASyncFuture[int]", other: int) -> "ASyncFuture[int]": ...
    @overload
    def __rsub__(self: "ASyncFuture[float]", other: float) -> "ASyncFuture[float]": ...
    @overload
    def __rsub__(self: "ASyncFuture[float]", other: int) -> "ASyncFuture[float]": ...
    @overload
    def __rsub__(self: "ASyncFuture[int]", other: float) -> "ASyncFuture[float]": ...
    @overload
    def __rsub__(
        self: "ASyncFuture[Decimal]", other: Decimal
    ) -> "ASyncFuture[Decimal]": ...
    @overload
    def __rsub__(
        self: "ASyncFuture[Decimal]", other: int
    ) -> "ASyncFuture[Decimal]": ...
    @overload
    def __rsub__(
        self: "ASyncFuture[int]", other: Decimal
    ) -> "ASyncFuture[Decimal]": ...
    @overload
    def __rsub__(
        self: "ASyncFuture[int]", other: Awaitable[int]
    ) -> "ASyncFuture[int]": ...
    @overload
    def __rsub__(
        self: "ASyncFuture[float]", other: Awaitable[float]
    ) -> "ASyncFuture[float]": ...
    @overload
    def __rsub__(
        self: "ASyncFuture[float]", other: Awaitable[int]
    ) -> "ASyncFuture[float]": ...
    @overload
    def __rsub__(
        self: "ASyncFuture[int]", other: Awaitable[float]
    ) -> "ASyncFuture[float]": ...
    @overload
    def __rsub__(
        self: "ASyncFuture[Decimal]", other: Awaitable[Decimal]
    ) -> "ASyncFuture[Decimal]": ...
    @overload
    def __rsub__(
        self: "ASyncFuture[Decimal]", other: Awaitable[int]
    ) -> "ASyncFuture[Decimal]": ...
    @overload
    def __rsub__(
        self: "ASyncFuture[int]", other: Awaitable[Decimal]
    ) -> "ASyncFuture[Decimal]": ...
    def __rsub__(self, other) -> "ASyncFuture":
        return ASyncFuture(
            self.__rsub(other), dependencies=self.__list_dependencies(other)
        )

    def __rmul__(self, other) -> "ASyncFuture":
        return ASyncFuture(
            self.__rmul(other), dependencies=self.__list_dependencies(other)
        )

    def __rtruediv__(self, other) -> "ASyncFuture":
        return ASyncFuture(
            self.__rtruediv(other), dependencies=self.__list_dependencies(other)
        )

    def __rfloordiv__(self, other) -> "ASyncFuture":
        return ASyncFuture(
            self.__rfloordiv(other), dependencies=self.__list_dependencies(other)
        )

    def __rpow__(self, other) -> "ASyncFuture":
        return ASyncFuture(
            self.__rpow(other), dependencies=self.__list_dependencies(other)
        )

    def __eq__(self, other) -> "ASyncFuture":
        return bool(
            ASyncFuture(self.__eq(other), dependencies=self.__list_dependencies(other))
        )

    def __gt__(self, other) -> "ASyncFuture":
        return ASyncFuture(
            self.__gt(other), dependencies=self.__list_dependencies(other)
        )

    def __ge__(self, other) -> "ASyncFuture":
        return ASyncFuture(
            self.__ge(other), dependencies=self.__list_dependencies(other)
        )

    def __lt__(self, other) -> "ASyncFuture":
        return ASyncFuture(
            self.__lt(other), dependencies=self.__list_dependencies(other)
        )

    def __le__(self, other) -> "ASyncFuture":
        return ASyncFuture(
            self.__le(other), dependencies=self.__list_dependencies(other)
        )

    # Maths

    @overload
    async def __add(self: "ASyncFuture[int]", other: int) -> "ASyncFuture[int]": ...
    @overload
    async def __add(
        self: "ASyncFuture[float]", other: float
    ) -> "ASyncFuture[float]": ...
    @overload
    async def __add(self: "ASyncFuture[float]", other: int) -> "ASyncFuture[float]": ...
    @overload
    async def __add(self: "ASyncFuture[int]", other: float) -> "ASyncFuture[float]": ...
    @overload
    async def __add(
        self: "ASyncFuture[Decimal]", other: Decimal
    ) -> "ASyncFuture[Decimal]": ...
    @overload
    async def __add(
        self: "ASyncFuture[Decimal]", other: int
    ) -> "ASyncFuture[Decimal]": ...
    @overload
    async def __add(
        self: "ASyncFuture[int]", other: Decimal
    ) -> "ASyncFuture[Decimal]": ...
    @overload
    async def __add(
        self: "ASyncFuture[int]", other: Awaitable[int]
    ) -> "ASyncFuture[int]": ...
    @overload
    async def __add(
        self: "ASyncFuture[float]", other: Awaitable[float]
    ) -> "ASyncFuture[float]": ...
    @overload
    async def __add(
        self: "ASyncFuture[float]", other: Awaitable[int]
    ) -> "ASyncFuture[float]": ...
    @overload
    async def __add(
        self: "ASyncFuture[int]", other: Awaitable[float]
    ) -> "ASyncFuture[float]": ...
    @overload
    async def __add(
        self: "ASyncFuture[Decimal]", other: Awaitable[Decimal]
    ) -> "ASyncFuture[Decimal]": ...
    @overload
    async def __add(
        self: "ASyncFuture[Decimal]", other: Awaitable[int]
    ) -> "ASyncFuture[Decimal]": ...
    @overload
    async def __add(
        self: "ASyncFuture[int]", other: Awaitable[Decimal]
    ) -> "ASyncFuture[Decimal]": ...
    async def __add(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(self, other)
        return a + b

    @overload
    async def __sub(self: "ASyncFuture[int]", other: int) -> "ASyncFuture[int]": ...
    @overload
    async def __sub(
        self: "ASyncFuture[float]", other: float
    ) -> "ASyncFuture[float]": ...
    @overload
    async def __sub(self: "ASyncFuture[float]", other: int) -> "ASyncFuture[float]": ...
    @overload
    async def __sub(self: "ASyncFuture[int]", other: float) -> "ASyncFuture[float]": ...
    @overload
    async def __sub(
        self: "ASyncFuture[Decimal]", other: Decimal
    ) -> "ASyncFuture[Decimal]": ...
    @overload
    async def __sub(
        self: "ASyncFuture[Decimal]", other: int
    ) -> "ASyncFuture[Decimal]": ...
    @overload
    async def __sub(
        self: "ASyncFuture[int]", other: Decimal
    ) -> "ASyncFuture[Decimal]": ...
    @overload
    async def __sub(
        self: "ASyncFuture[int]", other: Awaitable[int]
    ) -> "ASyncFuture[int]": ...
    @overload
    async def __sub(
        self: "ASyncFuture[float]", other: Awaitable[float]
    ) -> "ASyncFuture[float]": ...
    @overload
    async def __sub(
        self: "ASyncFuture[float]", other: Awaitable[int]
    ) -> "ASyncFuture[float]": ...
    @overload
    async def __sub(
        self: "ASyncFuture[int]", other: Awaitable[float]
    ) -> "ASyncFuture[float]": ...
    @overload
    async def __sub(
        self: "ASyncFuture[Decimal]", other: Awaitable[Decimal]
    ) -> "ASyncFuture[Decimal]": ...
    @overload
    async def __sub(
        self: "ASyncFuture[Decimal]", other: Awaitable[int]
    ) -> "ASyncFuture[Decimal]": ...
    @overload
    async def __sub(
        self: "ASyncFuture[int]", other: Awaitable[Decimal]
    ) -> "ASyncFuture[Decimal]": ...
    async def __sub(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(self, other)
        return a - b

    async def __mul(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(self, other)
        return a * b

    async def __truediv(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(self, other)
        return a / b

    async def __floordiv(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(self, other)
        return a // b

    async def __pow(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(self, other)
        return a**b

    # rMaths
    @overload
    async def __radd(self: "ASyncFuture[int]", other: int) -> "ASyncFuture[int]": ...
    @overload
    async def __radd(
        self: "ASyncFuture[float]", other: float
    ) -> "ASyncFuture[float]": ...
    @overload
    async def __radd(
        self: "ASyncFuture[float]", other: int
    ) -> "ASyncFuture[float]": ...
    @overload
    async def __radd(
        self: "ASyncFuture[int]", other: float
    ) -> "ASyncFuture[float]": ...
    @overload
    async def __radd(
        self: "ASyncFuture[Decimal]", other: Decimal
    ) -> "ASyncFuture[Decimal]": ...
    @overload
    async def __radd(
        self: "ASyncFuture[Decimal]", other: int
    ) -> "ASyncFuture[Decimal]": ...
    @overload
    async def __radd(
        self: "ASyncFuture[int]", other: Decimal
    ) -> "ASyncFuture[Decimal]": ...
    @overload
    async def __radd(
        self: "ASyncFuture[int]", other: Awaitable[int]
    ) -> "ASyncFuture[int]": ...
    @overload
    async def __radd(
        self: "ASyncFuture[float]", other: Awaitable[float]
    ) -> "ASyncFuture[float]": ...
    @overload
    async def __radd(
        self: "ASyncFuture[float]", other: Awaitable[int]
    ) -> "ASyncFuture[float]": ...
    @overload
    async def __radd(
        self: "ASyncFuture[int]", other: Awaitable[float]
    ) -> "ASyncFuture[float]": ...
    @overload
    async def __radd(
        self: "ASyncFuture[Decimal]", other: Awaitable[Decimal]
    ) -> "ASyncFuture[Decimal]": ...
    @overload
    async def __radd(
        self: "ASyncFuture[Decimal]", other: Awaitable[int]
    ) -> "ASyncFuture[Decimal]": ...
    @overload
    async def __radd(
        self: "ASyncFuture[int]", other: Awaitable[Decimal]
    ) -> "ASyncFuture[Decimal]": ...
    async def __radd(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(other, self)
        return a + b

    @overload
    async def __rsub(self: "ASyncFuture[int]", other: int) -> "ASyncFuture[int]": ...
    @overload
    async def __rsub(
        self: "ASyncFuture[float]", other: float
    ) -> "ASyncFuture[float]": ...
    @overload
    async def __rsub(
        self: "ASyncFuture[float]", other: int
    ) -> "ASyncFuture[float]": ...
    @overload
    async def __rsub(
        self: "ASyncFuture[int]", other: float
    ) -> "ASyncFuture[float]": ...
    @overload
    async def __rsub(
        self: "ASyncFuture[Decimal]", other: Decimal
    ) -> "ASyncFuture[Decimal]": ...
    @overload
    async def __rsub(
        self: "ASyncFuture[Decimal]", other: int
    ) -> "ASyncFuture[Decimal]": ...
    @overload
    async def __rsub(
        self: "ASyncFuture[int]", other: Decimal
    ) -> "ASyncFuture[Decimal]": ...
    @overload
    async def __rsub(
        self: "ASyncFuture[int]", other: Awaitable[int]
    ) -> "ASyncFuture[int]": ...
    @overload
    async def __rsub(
        self: "ASyncFuture[float]", other: Awaitable[float]
    ) -> "ASyncFuture[float]": ...
    @overload
    async def __rsub(
        self: "ASyncFuture[float]", other: Awaitable[int]
    ) -> "ASyncFuture[float]": ...
    @overload
    async def __rsub(
        self: "ASyncFuture[int]", other: Awaitable[float]
    ) -> "ASyncFuture[float]": ...
    @overload
    async def __rsub(
        self: "ASyncFuture[Decimal]", other: Awaitable[Decimal]
    ) -> "ASyncFuture[Decimal]": ...
    @overload
    async def __rsub(
        self: "ASyncFuture[Decimal]", other: Awaitable[int]
    ) -> "ASyncFuture[Decimal]": ...
    @overload
    async def __rsub(
        self: "ASyncFuture[int]", other: Awaitable[Decimal]
    ) -> "ASyncFuture[Decimal]": ...
    async def __rsub(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(other, self)
        return a - b

    async def __rmul(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(other, self)
        return a * b

    async def __rtruediv(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(other, self)
        return a / b

    async def __rfloordiv(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(other, self)
        return a // b

    async def __rpow(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(other, self)
        return a**b

    async def __iadd(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(self, other)
        self._result = a + b
        return self._result

    async def __isub(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(self, other)
        self._result = a - b
        return self._result

    async def __imul(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(self, other)
        self._result = a * b
        return self._result

    async def __itruediv(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(self, other)
        self._result = a / b
        return self._result

    async def __ifloordiv(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(self, other)
        self._result = a // b
        return self._result

    async def __ipow(self, other) -> "Any":
        a, b = await _gather_check_and_materialize(self, other)
        self._result = a**b
        return self._result

    # Comparisons
    async def __eq(self, other) -> bool:
        a, b = await _gather_check_and_materialize(self, other)
        return a == b

    async def __gt(self, other) -> bool:
        a, b = await _gather_check_and_materialize(self, other)
        return a > b

    async def __ge(self, other) -> bool:
        a, b = await _gather_check_and_materialize(self, other)
        return a >= b

    async def __lt(self, other) -> bool:
        a, b = await _gather_check_and_materialize(self, other)
        return a < b

    async def __le(self, other) -> bool:
        a, b = await _gather_check_and_materialize(self, other)
        return a <= b

    # not sure what to call these
    async def __contains(self, item: Any) -> bool:
        _self, _item = await _gather_check_and_materialize(self, item)
        return _item in _self

    # conversion
    # NOTE: We aren't allowed to return ASyncFutures here :(
    def __bool__(self) -> bool:
        return bool(_materialize(self))

    def __bytes__(self) -> bytes:
        return bytes(_materialize(self))

    def __str__(self) -> str:
        return str(_materialize(self))

    def __int__(self) -> int:
        return int(_materialize(self))

    def __float__(self) -> float:
        return float(_materialize(self))

    # WIP internals

    @property
    def __dependants__(self) -> Set["ASyncFuture"]:
        """
        Returns the set of dependants for this ASyncFuture, including nested dependants.
        """
        dependants = set()
        for dep in self.__dependants:
            dependants.add(dep)
            dependants.union(dep.__dependants__)
        return dependants

    @property
    def __dependencies__(self) -> Set["ASyncFuture"]:
        """
        Returns the set of dependencies for this ASyncFuture, including nested dependencies.
        """
        dependencies = set()
        for dep in self.__dependencies:
            dependencies.add(dep)
            dependencies.union(dep.__dependencies__)
        return dependencies

    def __sizeof__(self) -> int:
        if isinstance(self.__awaitable__, Coroutine):
            return sum(
                sys.getsizeof(v) for v in self.__awaitable__.cr_frame.f_locals.values()
            )
        elif isinstance(self.__awaitable__, asyncio.Future):
            raise NotImplementedError
        raise NotImplementedError


@final
class _ASyncFutureWrappedFn(Callable[P, ASyncFuture[T]]):
    """
    A callable class to wrap functions and return ASyncFuture objects.
    """

    __slots__ = "callable", "wrapped", "_callable_name"

    def __init__(
        self,
        callable: AnyFn[P, T] = None,
        **kwargs: Unpack[ModifierKwargs],
    ):
        from a_sync import a_sync

        if callable:
            self.callable = callable
            """The callable function."""

            self._callable_name = callable.__name__
            """The name of the callable function."""

            a_sync_callable = a_sync(callable, default="async", **kwargs)

            @wraps(callable)
            def future_wrap(*args: P.args, **kwargs: P.kwargs) -> "ASyncFuture[T]":
                return ASyncFuture(a_sync_callable(*args, **kwargs, sync=False))

            self.wrapped = future_wrap
            """The wrapped function returning ASyncFuture."""
        else:
            self.wrapped = partial(_ASyncFutureWrappedFn, **kwargs)

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> ASyncFuture[T]:
        return self.wrapped(*args, **kwargs)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.callable}>"

    def __get__(
        self, instance: I, owner: Type[I]
    ) -> Union[Self, "_ASyncFutureInstanceMethod[I, P, T]"]:
        if owner is None:
            return self
        else:
            return _ASyncFutureInstanceMethod(self, instance)


@final
class _ASyncFutureInstanceMethod(Generic[I, P, T]):
    # NOTE: probably could just replace this with functools.partial
    """
    A class to handle instance methods wrapped as ASyncFuture.
    """

    __module__: str
    """The module name of the wrapper."""

    __name__: str
    """The name of the wrapper."""

    __qualname__: str
    """The qualified name of the wrapper."""

    __doc__: Optional[str]
    """The docstring of the wrapper."""

    __annotations__: Dict[str, Any]
    """The annotations of the wrapper."""

    __instance: I
    """The instance to which the method is bound."""

    __wrapper: _ASyncFutureWrappedFn[P, T]
    """The wrapper function."""

    def __init__(
        self,
        wrapper: _ASyncFutureWrappedFn[P, T],
        instance: I,
    ) -> None:
        try:
            self.__module__ = wrapper.__module__
        except AttributeError:
            pass
        try:
            self.__name__ = wrapper.__name__
        except AttributeError:
            pass
        try:
            self.__qualname__ = wrapper.__qualname__
        except AttributeError:
            pass
        try:
            self.__doc__ = wrapper.__doc__
        except AttributeError:
            pass
        try:
            self.__annotations__ = wrapper.__annotations__
        except AttributeError:
            pass
        try:
            self.__dict__.update(wrapper.__dict__)
        except AttributeError:
            pass
        self.__instance = instance
        self.__wrapper = wrapper

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} for {self.__wrapper.callable} bound to {self.__instance}>"

    def __call__(self, /, *fn_args: P.args, **fn_kwargs: P.kwargs) -> T:
        return self.__wrapper(self.__instance, *fn_args, **fn_kwargs)


__all__ = ["future", "ASyncFuture"]



import functools
import inspect
import logging
import sys

from async_lru import _LRUCacheWrapper
from async_property.base import AsyncPropertyDescriptor  # type: ignore [import]
from async_property.cached import AsyncCachedPropertyDescriptor  # type: ignore [import]

from a_sync._typing import *
from a_sync.a_sync import _flags, _helpers, _kwargs
from a_sync.a_sync.modifiers.manager import ModifierManager

if TYPE_CHECKING:
    from a_sync import TaskMapping
    from a_sync.a_sync.method import (
        ASyncBoundMethod,
        ASyncBoundMethodAsyncDefault,
        ASyncBoundMethodSyncDefault,
    )

logger = logging.getLogger(__name__)


class _ModifiedMixin:
    """
    A mixin class for internal use that provides functionality for applying modifiers to functions.

    This class is used as a base for :class:`~ASyncFunction` and its variants, such as
    `ASyncFunctionAsyncDefault` and `ASyncFunctionSyncDefault`, to handle the application
    of async and sync modifiers to functions. Modifiers can alter the behavior of functions,
    such as converting sync functions to async, applying caching, or rate limiting.
    """

    modifiers: ModifierManager
    __slots__ = "modifiers", "wrapped"

    def _asyncify(self, func: SyncFn[P, T]) -> CoroFn[P, T]:
        """
        Converts a synchronous function to an asynchronous one and applies async modifiers.

        Args:
            func: The synchronous function to be converted.

        Returns:
            The asynchronous version of the function with applied modifiers.
        """
        coro_fn = _helpers._asyncify(func, self.modifiers.executor)
        return self.modifiers.apply_async_modifiers(coro_fn)

    @functools.cached_property
    def _await(self) -> Callable[[Awaitable[T]], T]:
        """
        Applies sync modifiers to the _helpers._await function and caches it.

        Returns:
            The modified _await function.
        """
        return self.modifiers.apply_sync_modifiers(_helpers._await)

    @functools.cached_property
    def default(self) -> DefaultMode:
        """
        Gets the default execution mode (sync, async, or None) for the function.

        Returns:
            The default execution mode.
        """
        return self.modifiers.default


def _validate_wrapped_fn(fn: Callable) -> None:
    """Ensures 'fn' is an appropriate function for wrapping with a_sync.

    Args:
        fn: The function to validate.

    Raises:
        TypeError: If the input is not callable.
        RuntimeError: If the function has arguments with names that conflict with viable flags.
    """
    if isinstance(fn, (AsyncPropertyDescriptor, AsyncCachedPropertyDescriptor)):
        return  # These are always valid
    if not callable(fn):
        raise TypeError(f"Input is not callable. Unable to decorate {fn}")
    if isinstance(fn, _LRUCacheWrapper):
        fn = fn.__wrapped__
    _check_not_genfunc(fn)
    fn_args = inspect.getfullargspec(fn)[0]
    for flag in _flags.VIABLE_FLAGS:
        if flag in fn_args:
            raise RuntimeError(
                f"{fn} must not have any arguments with the following names: {_flags.VIABLE_FLAGS}"
            )


class ASyncFunction(_ModifiedMixin, Generic[P, T]):
    """
    A callable wrapper object that can be executed both synchronously and asynchronously.

    This class wraps a function or coroutine function, allowing it to be called in both
    synchronous and asynchronous contexts. It provides a flexible interface for handling
    different execution modes and applying various modifiers to the function's behavior.

    The class supports various modifiers that can alter the behavior of the function,
    such as caching, rate limiting, and execution in specific contexts (e.g., thread pools).

    Example:
        async def my_coroutine(x: int) -> str:
            return str(x)

        func = ASyncFunction(my_coroutine)

        # Synchronous call
        result = func(5, sync=True)  # returns "5"

        # Asynchronous call
        result = await func(5)  # returns "5"
    """

    # NOTE: We can't use __slots__ here because it breaks functools.update_wrapper

    @overload
    def __init__(self, fn: CoroFn[P, T], **modifiers: Unpack[ModifierKwargs]) -> None:
        ...
        # TODO write specific docs for this overload

    @overload
    def __init__(self, fn: SyncFn[P, T], **modifiers: Unpack[ModifierKwargs]) -> None:
        ...
        # TODO write specific docs for this overload

    def __init__(self, fn: AnyFn[P, T], **modifiers: Unpack[ModifierKwargs]) -> None:
        """
        Initializes an ASyncFunction instance.

        Args:
            fn: The function to wrap.
            **modifiers: Keyword arguments for function modifiers.
        """
        _validate_wrapped_fn(fn)

        self.modifiers = ModifierManager(modifiers)
        """A :class:`~ModifierManager` instance managing function modifiers."""

        self.__wrapped__ = fn
        """The original function that was wrapped."""

        functools.update_wrapper(self, self.__wrapped__)
        if self.__doc__ is None:
            self.__doc__ = f"Since `{self.__name__}` is an {self.__docstring_append__}"
        else:
            self.__doc__ += (
                f"\n\nSince `{self.__name__}` is an {self.__docstring_append__}"
            )

    @overload
    def __call__(self, *args: P.args, sync: Literal[True], **kwargs: P.kwargs) -> T:
        ...
        # TODO write specific docs for this overload

    @overload
    def __call__(
        self, *args: P.args, sync: Literal[False], **kwargs: P.kwargs
    ) -> Coroutine[Any, Any, T]:
        ...
        # TODO write specific docs for this overload

    @overload
    def __call__(
        self, *args: P.args, asynchronous: Literal[False], **kwargs: P.kwargs
    ) -> T:
        ...
        # TODO write specific docs for this overload

    @overload
    def __call__(
        self, *args: P.args, asynchronous: Literal[True], **kwargs: P.kwargs
    ) -> Coroutine[Any, Any, T]:
        ...
        # TODO write specific docs for this overload

    @overload
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> MaybeCoro[T]:
        ...
        # TODO write specific docs for this overload

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> MaybeCoro[T]:
        """
        Calls the wrapped function either synchronously or asynchronously.

        This method determines whether to execute the wrapped function synchronously
        or asynchronously based on the default mode and any provided flags.

        Args:
            *args: Positional arguments to pass to the wrapped function.
            **kwargs: Keyword arguments to pass to the wrapped function.

        Raises:
            Exception: Any exception that may be raised by the wrapped function.
        """
        logger.debug(
            "calling %s fn: %s with args: %s kwargs: %s", self, self.fn, args, kwargs
        )
        return self.fn(*args, **kwargs)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.__module__}.{self.__name__} at {hex(id(self))}>"

    @functools.cached_property
    def fn(
        self,
    ):  # -> Union[SyncFn[[CoroFn[P, T]], MaybeAwaitable[T]], SyncFn[[SyncFn[P, T]], MaybeAwaitable[T]]]:
        """
        Returns the final wrapped version of :attr:`ASyncFunction._fn` decorated with all of the a_sync goodness.

        Returns:
            The final wrapped function.
        """
        return self._async_wrap if self._async_def else self._sync_wrap

    if sys.version_info >= (3, 11) or TYPE_CHECKING:
        # we can specify P.args in python>=3.11 but in lower versions it causes a crash. Everything should still type check correctly on all versions.
        def map(
            self,
            *iterables: AnyIterable[P.args],
            concurrency: Optional[int] = None,
            task_name: str = "",
            **function_kwargs: P.kwargs,
        ) -> "TaskMapping[P, T]":
            """
            Creates a TaskMapping for the wrapped function with the given iterables.

            Args:
                *iterables: Iterable objects to be used as arguments for the function.
                concurrency: Optional maximum number of concurrent tasks.
                task_name: Optional name for the tasks.
                **function_kwargs: Additional keyword arguments to pass to the function.

            Returns:
                A TaskMapping object for managing concurrent execution.
            """
            from a_sync import TaskMapping

            return TaskMapping(
                self,
                *iterables,
                concurrency=concurrency,
                name=task_name,
                **function_kwargs,
            )

        async def any(
            self,
            *iterables: AnyIterable[P.args],
            concurrency: Optional[int] = None,
            task_name: str = "",
            **function_kwargs: P.kwargs,
        ) -> bool:
            """
            Checks if any result of the function applied to the iterables is truthy.

            Args:
                *iterables: Iterable objects to be used as arguments for the function.
                concurrency: Optional maximum number of concurrent tasks.
                task_name: Optional name for the tasks.
                **function_kwargs: Additional keyword arguments to pass to the function.

            Returns:
                True if any result is truthy, otherwise False.
            """
            return await self.map(
                *iterables,
                concurrency=concurrency,
                task_name=task_name,
                **function_kwargs,
            ).any(pop=True, sync=False)

        async def all(
            self,
            *iterables: AnyIterable[P.args],
            concurrency: Optional[int] = None,
            task_name: str = "",
            **function_kwargs: P.kwargs,
        ) -> bool:
            """
            Checks if all results of the function applied to the iterables are truthy.

            Args:
                *iterables: Iterable objects to be used as arguments for the function.
                concurrency: Optional maximum number of concurrent tasks.
                task_name: Optional name for the tasks.
                **function_kwargs: Additional keyword arguments to pass to the function.

            Returns:
                True if all results are truthy, otherwise False.
            """
            return await self.map(
                *iterables,
                concurrency=concurrency,
                task_name=task_name,
                **function_kwargs,
            ).all(pop=True, sync=False)

        async def min(
            self,
            *iterables: AnyIterable[P.args],
            concurrency: Optional[int] = None,
            task_name: str = "",
            **function_kwargs: P.kwargs,
        ) -> T:
            """
            Finds the minimum result of the function applied to the iterables.

            Args:
                *iterables: Iterable objects to be used as arguments for the function.
                concurrency: Optional maximum number of concurrent tasks.
                task_name: Optional name for the tasks.
                **function_kwargs: Additional keyword arguments to pass to the function.

            Returns:
                The minimum result.
            """
            return await self.map(
                *iterables,
                concurrency=concurrency,
                task_name=task_name,
                **function_kwargs,
            ).min(pop=True, sync=False)

        async def max(
            self,
            *iterables: AnyIterable[P.args],
            concurrency: Optional[int] = None,
            task_name: str = "",
            **function_kwargs: P.kwargs,
        ) -> T:
            """
            Finds the maximum result of the function applied to the iterables.

            Args:
                *iterables: Iterable objects to be used as arguments for the function.
                concurrency: Optional maximum number of concurrent tasks.
                task_name: Optional name for the tasks.
                **function_kwargs: Additional keyword arguments to pass to the function.

            Returns:
                The maximum result.
            """
            return await self.map(
                *iterables,
                concurrency=concurrency,
                task_name=task_name,
                **function_kwargs,
            ).max(pop=True, sync=False)

        async def sum(
            self,
            *iterables: AnyIterable[P.args],
            concurrency: Optional[int] = None,
            task_name: str = "",
            **function_kwargs: P.kwargs,
        ) -> T:
            """
            Calculates the sum of the results of the function applied to the iterables.

            Args:
                *iterables: Iterable objects to be used as arguments for the function.
                concurrency: Optional maximum number of concurrent tasks.
                task_name: Optional name for the tasks.
                **function_kwargs: Additional keyword arguments to pass to the function.

            Returns:
                The sum of the results.
            """
            return await self.map(
                *iterables,
                concurrency=concurrency,
                task_name=task_name,
                **function_kwargs,
            ).sum(pop=True, sync=False)

    else:

        def map(
            self,
            *iterables: AnyIterable[Any],
            concurrency: Optional[int] = None,
            task_name: str = "",
            **function_kwargs: P.kwargs,
        ) -> "TaskMapping[P, T]":
            """
            Creates a TaskMapping for the wrapped function with the given iterables.

            Args:
                *iterables: Iterable objects to be used as arguments for the function.
                concurrency: Optional maximum number of concurrent tasks.
                task_name: Optional name for the tasks.
                **function_kwargs: Additional keyword arguments to pass to the function.

            Returns:
                A TaskMapping object for managing concurrent execution.
            """
            from a_sync import TaskMapping

            return TaskMapping(
                self,
                *iterables,
                concurrency=concurrency,
                name=task_name,
                **function_kwargs,
            )

        async def any(
            self,
            *iterables: AnyIterable[Any],
            concurrency: Optional[int] = None,
            task_name: str = "",
            **function_kwargs: P.kwargs,
        ) -> bool:
            """
            Checks if any result of the function applied to the iterables is truthy.

            Args:
                *iterables: Iterable objects to be used as arguments for the function.
                concurrency: Optional maximum number of concurrent tasks.
                task_name: Optional name for the tasks.
                **function_kwargs: Additional keyword arguments to pass to the function.

            Returns:
                True if any result is truthy, otherwise False.
            """
            return await self.map(
                *iterables,
                concurrency=concurrency,
                task_name=task_name,
                **function_kwargs,
            ).any(pop=True, sync=False)

        async def all(
            self,
            *iterables: AnyIterable[Any],
            concurrency: Optional[int] = None,
            task_name: str = "",
            **function_kwargs: P.kwargs,
        ) -> bool:
            """
            Checks if all results of the function applied to the iterables are truthy.

            Args:
                *iterables: Iterable objects to be used as arguments for the function.
                concurrency: Optional maximum number of concurrent tasks.
                task_name: Optional name for the tasks.
                **function_kwargs: Additional keyword arguments to pass to the function.

            Returns:
                True if all results are truthy, otherwise False.
            """
            return await self.map(
                *iterables,
                concurrency=concurrency,
                task_name=task_name,
                **function_kwargs,
            ).all(pop=True, sync=False)

        async def min(
            self,
            *iterables: AnyIterable[Any],
            concurrency: Optional[int] = None,
            task_name: str = "",
            **function_kwargs: P.kwargs,
        ) -> T:
            """
            Finds the minimum result of the function applied to the iterables.

            Args:
                *iterables: Iterable objects to be used as arguments for the function.
                concurrency: Optional maximum number of concurrent tasks.
                task_name: Optional name for the tasks.
                **function_kwargs: Additional keyword arguments to pass to the function.

            Returns:
                The minimum result.
            """
            return await self.map(
                *iterables,
                concurrency=concurrency,
                task_name=task_name,
                **function_kwargs,
            ).min(pop=True, sync=False)

        async def max(
            self,
            *iterables: AnyIterable[Any],
            concurrency: Optional[int] = None,
            task_name: str = "",
            **function_kwargs: P.kwargs,
        ) -> T:
            """
            Finds the maximum result of the function applied to the iterables.

            Args:
                *iterables: Iterable objects to be used as arguments for the function.
                concurrency: Optional maximum number of concurrent tasks.
                task_name: Optional name for the tasks.
                **function_kwargs: Additional keyword arguments to pass to the function.

            Returns:
                The maximum result.
            """
            return await self.map(
                *iterables,
                concurrency=concurrency,
                task_name=task_name,
                **function_kwargs,
            ).max(pop=True, sync=False)

        async def sum(
            self,
            *iterables: AnyIterable[Any],
            concurrency: Optional[int] = None,
            task_name: str = "",
            **function_kwargs: P.kwargs,
        ) -> T:
            """
            Calculates the sum of the results of the function applied to the iterables.

            Args:
                *iterables: Iterable objects to be used as arguments for the function.
                concurrency: Optional maximum number of concurrent tasks.
                task_name: Optional name for the tasks.
                **function_kwargs: Additional keyword arguments to pass to the function.

            Returns:
                The sum of the results.
            """
            return await self.map(
                *iterables,
                concurrency=concurrency,
                task_name=task_name,
                **function_kwargs,
            ).sum(pop=True, sync=False)

    @functools.cached_property
    def _sync_default(self) -> bool:
        """
        Determines the default execution mode (sync or async) for the function.

        If the user did not specify a default, this method defers to the function's
        definition (sync vs async def).

        Returns:
            True if the default is sync, False if async.
        """
        return (
            True
            if self.default == "sync"
            else False if self.default == "async" else not self._async_def
        )

    @functools.cached_property
    def _async_def(self) -> bool:
        """
        Checks if the wrapped function is an asynchronous function.

        Returns:
            True if the function is asynchronous, otherwise False.
        """
        return asyncio.iscoroutinefunction(self.__wrapped__)

    def _run_sync(self, kwargs: dict) -> bool:
        """
        Determines whether to run the function synchronously or asynchronously.

        This method checks for a flag in the kwargs and defers to it if present.
        If no flag is specified, it defers to the default execution mode.

        Args:
            kwargs: The keyword arguments passed to the function.

        Returns:
            True if the function should run synchronously, otherwise False.
        """
        if flag := _kwargs.get_flag_name(kwargs):
            # If a flag was specified in the kwargs, we will defer to it.
            return _kwargs.is_sync(flag, kwargs, pop_flag=True)
        else:
            # No flag specified in the kwargs, we will defer to 'default'.
            return self._sync_default

    @functools.cached_property
    def _asyncified(self) -> CoroFn[P, T]:
        """
        Converts the wrapped function to an asynchronous function and applies both sync and async modifiers.

        Raises:
            TypeError: If the wrapped function is already asynchronous.

        Returns:
            The asynchronous version of the wrapped function.
        """
        if self._async_def:
            raise TypeError(
                f"Can only be applied to sync functions, not {self.__wrapped__}"
            )
        return self._asyncify(self._modified_fn)  # type: ignore [arg-type]

    @functools.cached_property
    def _modified_fn(self) -> AnyFn[P, T]:
        """
        Applies modifiers to the wrapped function.

        If the wrapped function is an asynchronous function, this method applies async modifiers.
        If the wrapped function is a synchronous function, this method applies sync modifiers.

        Returns:
            The modified function.
        """
        if self._async_def:
            return self.modifiers.apply_async_modifiers(self.__wrapped__)  # type: ignore [arg-type]
        return self.modifiers.apply_sync_modifiers(self.__wrapped__)  # type: ignore [return-value]

    @functools.cached_property
    def _async_wrap(self):  # -> SyncFn[[CoroFn[P, T]], MaybeAwaitable[T]]:
        """
        The final wrapper if the wrapped function is an asynchronous function.

        This method applies the appropriate modifiers and determines whether to await the result.

        Returns:
            The wrapped function with async handling.
        """

        @functools.wraps(self._modified_fn)
        def async_wrap(*args: P.args, **kwargs: P.kwargs) -> MaybeAwaitable[T]:  # type: ignore [name-defined]
            should_await = self._run_sync(
                kwargs
            )  # Must take place before coro is created, we're popping a kwarg.
            coro = self._modified_fn(*args, **kwargs)
            return self._await(coro) if should_await else coro

        return async_wrap

    @functools.cached_property
    def _sync_wrap(self):  # -> SyncFn[[SyncFn[P, T]], MaybeAwaitable[T]]:
        """
        The final wrapper if the wrapped function is a synchronous function.

        This method applies the appropriate modifiers and determines whether to run the function synchronously or asynchronously.

        Returns:
            The wrapped function with sync handling.
        """

        @functools.wraps(self._modified_fn)
        def sync_wrap(*args: P.args, **kwargs: P.kwargs) -> MaybeAwaitable[T]:  # type: ignore [name-defined]
            if self._run_sync(kwargs):
                return self._modified_fn(*args, **kwargs)
            return self._asyncified(*args, **kwargs)

        return sync_wrap

    __docstring_append__ = ":class:`~a_sync.a_sync.function.ASyncFunction`, you can optionally pass either a `sync` or `asynchronous` kwarg with a boolean value."


if sys.version_info < (3, 10):
    _inherit = ASyncFunction[AnyFn[P, T], ASyncFunction[P, T]]
else:
    _inherit = ASyncFunction[[AnyFn[P, T]], ASyncFunction[P, T]]


class ASyncDecorator(_ModifiedMixin):
    def __init__(self, **modifiers: Unpack[ModifierKwargs]) -> None:
        """
        Initializes an ASyncDecorator instance.

        Args:
            **modifiers: Keyword arguments for function modifiers.

        Raises:
            ValueError: If 'default' is not 'sync', 'async', or None.
        """
        assert "default" in modifiers, modifiers
        self.modifiers = ModifierManager(modifiers)
        self.validate_inputs()

    def validate_inputs(self) -> None:
        """
        Validates the input modifiers.

        Raises:
            ValueError: If 'default' is not 'sync', 'async', or None.
        """
        if self.modifiers.default not in ["sync", "async", None]:
            raise ValueError(
                f"'default' must be either 'sync', 'async', or None. You passed {self.modifiers.default}."
            )

    @overload
    def __call__(self, func: AnyFn[Concatenate[B, P], T]) -> "ASyncBoundMethod[B, P, T]":  # type: ignore [override]
        ...  # TODO write specific docs for this overload

    @overload
    def __call__(self, func: AnyFn[P, T]) -> ASyncFunction[P, T]:  # type: ignore [override]
        ...  # TODO write specific docs for this overload

    def __call__(self, func: AnyFn[P, T]) -> ASyncFunction[P, T]:  # type: ignore [override]
        """
        Decorates a function with async or sync behavior based on the default modifier.

        Args:
            func: The function to decorate.

        Returns:
            An ASyncFunction instance with the appropriate default behavior.
        """
        if self.default == "async":
            return ASyncFunctionAsyncDefault(func, **self.modifiers)
        elif self.default == "sync":
            return ASyncFunctionSyncDefault(func, **self.modifiers)
        elif asyncio.iscoroutinefunction(func):
            return ASyncFunctionAsyncDefault(func, **self.modifiers)
        else:
            return ASyncFunctionSyncDefault(func, **self.modifiers)


def _check_not_genfunc(func: Callable) -> None:
    """Raises an error if the function is a generator or async generator.

    Args:
        func: The function to check.

    Raises:
        ValueError: If the function is a generator or async generator.
    """
    if inspect.isasyncgenfunction(func) or inspect.isgeneratorfunction(func):
        raise ValueError("unable to decorate generator functions with this decorator")


# Mypy helper classes


class ASyncFunctionSyncDefault(ASyncFunction[P, T]):
    """A specialized :class:`~ASyncFunction` that defaults to synchronous execution.

    This class is used when the :func:`~a_sync` decorator is applied with `default='sync'`.
    It provides type hints to indicate that the default call behavior is synchronous and
    supports IDE type checking for most use cases.

    The wrapped function can still be called asynchronously by passing `sync=False`
    or `asynchronous=True` as a keyword argument.

    Example:
        @a_sync(default='sync')
        async def my_function(x: int) -> str:
            return str(x)

        # Synchronous call (default behavior)
        result = my_function(5)  # returns "5"

        # Asynchronous call
        result = await my_function(5, sync=False)  # returns "5"
    """

    @overload
    def __call__(
        self, *args: P.args, sync: Literal[True], **kwargs: P.kwargs
    ) -> T: ...  # TODO write specific docs for this overload
    @overload
    def __call__(
        self, *args: P.args, sync: Literal[False], **kwargs: P.kwargs
    ) -> Coroutine[Any, Any, T]: ...  # TODO write specific docs for this overload
    @overload
    def __call__(
        self, *args: P.args, asynchronous: Literal[False], **kwargs: P.kwargs
    ) -> T: ...  # TODO write specific docs for this overload
    @overload
    def __call__(
        self, *args: P.args, asynchronous: Literal[True], **kwargs: P.kwargs
    ) -> Coroutine[Any, Any, T]: ...  # TODO write specific docs for this overload
    @overload
    def __call__(
        self, *args: P.args, **kwargs: P.kwargs
    ) -> T: ...  # TODO write specific docs for this overload
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> MaybeCoro[T]:
        """Calls the wrapped function, defaulting to synchronous execution.

        This method overrides the base :meth:`ASyncFunction.__call__` to provide a synchronous
        default behavior.

        Args:
            *args: Positional arguments to pass to the wrapped function.
            **kwargs: Keyword arguments to pass to the wrapped function.

        Raises:
            Exception: Any exception that may be raised by the wrapped function.

        Returns:
            The result of the function call.
        """
        return self.fn(*args, **kwargs)

    __docstring_append__ = ":class:`~a_sync.a_sync.function.ASyncFunctionSyncDefault`, you can optionally pass `sync=False` or `asynchronous=True` to force it to return a coroutine. Without either kwarg, it will run synchronously."


class ASyncFunctionAsyncDefault(ASyncFunction[P, T]):
    """
    A specialized :class:`~ASyncFunction` that defaults to asynchronous execution.

    This class is used when the :func:`~a_sync` decorator is applied with `default='async'`.
    It provides type hints to indicate that the default call behavior is asynchronous
    and supports IDE type checking for most use cases.

    The wrapped function can still be called synchronously by passing `sync=True`
    or `asynchronous=False` as a keyword argument.

    Example:
        @a_sync(default='async')
        async def my_function(x: int) -> str:
            return str(x)

        # Asynchronous call (default behavior)
        result = await my_function(5)  # returns "5"

        # Synchronous call
        result = my_function(5, sync=True)  # returns "5"
    """

    @overload
    def __call__(self, *args: P.args, sync: Literal[True], **kwargs: P.kwargs) -> T: ...

    # TODO write specific docs for this overload
    @overload
    def __call__(
        self, *args: P.args, sync: Literal[False], **kwargs: P.kwargs
    ) -> Coroutine[Any, Any, T]: ...

    # TODO write specific docs for this overload
    @overload
    def __call__(
        self, *args: P.args, asynchronous: Literal[False], **kwargs: P.kwargs
    ) -> T: ...

    # TODO write specific docs for this overload
    @overload
    def __call__(
        self, *args: P.args, asynchronous: Literal[True], **kwargs: P.kwargs
    ) -> Coroutine[Any, Any, T]: ...

    # TODO write specific docs for this overload
    @overload
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> Coroutine[Any, Any, T]: ...

    # TODO write specific docs for this overload
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> MaybeCoro[T]:
        """Calls the wrapped function, defaulting to asynchronous execution.

        This method overrides the base :meth:`ASyncFunction.__call__` to provide an asynchronous
        default behavior.

        Args:
            *args: Positional arguments to pass to the wrapped function.
            **kwargs: Keyword arguments to pass to the wrapped function.

        Raises:
            Exception: Any exception that may be raised by the wrapped function.

        Returns:
            The result of the function call.
        """
        return self.fn(*args, **kwargs)

    __docstring_append__ = ":class:`~a_sync.a_sync.function.ASyncFunctionAsyncDefault`, you can optionally pass `sync=True` or `asynchronous=False` to force it to run synchronously and return a value. Without either kwarg, it will return a coroutine for you to await."


class ASyncDecoratorSyncDefault(ASyncDecorator):
    @overload
    def __call__(self, func: AnyFn[Concatenate[B, P], T]) -> "ASyncBoundMethodSyncDefault[P, T]":  # type: ignore [override]
        ...
        # TODO write specific docs for this overload

    @overload
    def __call__(self, func: AnyBoundMethod[P, T]) -> ASyncFunctionSyncDefault[P, T]:  # type: ignore [override]
        ...  # TODO write specific docs for this overload

    @overload
    def __call__(self, func: AnyFn[P, T]) -> ASyncFunctionSyncDefault[P, T]:  # type: ignore [override]
        ...  # TODO write specific docs for this overload

    def __call__(self, func: AnyFn[P, T]) -> ASyncFunctionSyncDefault[P, T]:
        ...  # TODO write specific docs for this overload
        return ASyncFunctionSyncDefault(func, **self.modifiers)


class ASyncDecoratorAsyncDefault(ASyncDecorator):
    @overload
    def __call__(self, func: AnyFn[Concatenate[B, P], T]) -> "ASyncBoundMethodAsyncDefault[P, T]":  # type: ignore [override]
        ...  # TODO write specific docs for this overload

    @overload
    def __call__(self, func: AnyBoundMethod[P, T]) -> ASyncFunctionAsyncDefault[P, T]:  # type: ignore [override]
        ...  # TODO write specific docs for this overload

    @overload
    def __call__(self, func: AnyFn[P, T]) -> ASyncFunctionAsyncDefault[P, T]:  # type: ignore [override]
        ...  # TODO write specific docs for this overload

    def __call__(self, func: AnyFn[P, T]) -> ASyncFunctionAsyncDefault[P, T]:
        # TODO write specific docs for this overload
        return ASyncFunctionAsyncDefault(func, **self.modifiers)



"""
This module provides priority-based semaphore implementations. These semaphores allow 
waiters to be assigned priorities, ensuring that higher priority waiters are 
processed before lower priority ones.
"""

import asyncio
import heapq
import logging
from collections import deque
from functools import cached_property

from a_sync._typing import *
from a_sync.primitives.locks.semaphore import Semaphore

logger = logging.getLogger(__name__)


class Priority(Protocol):
    def __lt__(self, other) -> bool: ...


PT = TypeVar("PT", bound=Priority)

CM = TypeVar("CM", bound="_AbstractPrioritySemaphoreContextManager[Priority]")


class _AbstractPrioritySemaphore(Semaphore, Generic[PT, CM]):
    """
    A semaphore that allows prioritization of waiters.

    This semaphore manages waiters with associated priorities, ensuring that waiters with higher
    priorities are processed before those with lower priorities. If no priority is specified,
    the semaphore uses a default top priority.
    """

    name: Optional[str]
    _value: int
    _waiters: List["_AbstractPrioritySemaphoreContextManager[PT]"]  # type: ignore [assignment]
    _context_managers: Dict[PT, "_AbstractPrioritySemaphoreContextManager[PT]"]
    __slots__ = (
        "name",
        "_value",
        "_waiters",
        "_context_managers",
        "_capacity",
        "_potential_lost_waiters",
    )

    @property
    def _context_manager_class(
        self,
    ) -> Type["_AbstractPrioritySemaphoreContextManager[PT]"]:
        raise NotImplementedError

    @property
    def _top_priority(self) -> PT:
        # You can use this so you can set priorities with non numeric comparable values
        raise NotImplementedError

    def __init__(self, value: int = 1, *, name: Optional[str] = None) -> None:
        """Initializes the priority semaphore.

        Args:
            value: The initial capacity of the semaphore.
            name: An optional name for the semaphore, used for debugging.
        """

        self._context_managers = {}
        """A dictionary mapping priorities to their context managers."""

        self._capacity = value
        """The initial capacity of the semaphore."""

        super().__init__(value, name=name)
        self._waiters = []
        """A heap queue of context managers, sorted by priority."""

        # NOTE: This should (hopefully) be temporary
        self._potential_lost_waiters: List["asyncio.Future[None]"] = []
        """A list of futures representing waiters that might have been lost."""

    def __repr__(self) -> str:
        """Returns a string representation of the semaphore."""
        return f"<{self.__class__.__name__} name={self.name} capacity={self._capacity} value={self._value} waiters={self._count_waiters()}>"

    async def __aenter__(self) -> None:
        """Enters the semaphore context, acquiring it with the top priority."""
        await self[self._top_priority].acquire()

    async def __aexit__(self, *_) -> None:
        """Exits the semaphore context, releasing it with the top priority."""
        self[self._top_priority].release()

    async def acquire(self) -> Literal[True]:
        """Acquires the semaphore with the top priority."""
        return await self[self._top_priority].acquire()

    def __getitem__(
        self, priority: Optional[PT]
    ) -> "_AbstractPrioritySemaphoreContextManager[PT]":
        """Gets the context manager for a given priority.

        Args:
            priority: The priority for which to get the context manager. If None, uses the top priority.

        Returns:
            The context manager associated with the given priority.
        """
        priority = self._top_priority if priority is None else priority
        if priority not in self._context_managers:
            context_manager = self._context_manager_class(
                self, priority, name=self.name
            )
            heapq.heappush(self._waiters, context_manager)  # type: ignore [misc]
            self._context_managers[priority] = context_manager
        return self._context_managers[priority]

    def locked(self) -> bool:
        """Checks if the semaphore is locked.

        Returns:
            True if the semaphore cannot be acquired immediately, False otherwise.
        """
        return self._value == 0 or (
            any(
                cm._waiters and any(not w.cancelled() for w in cm._waiters)
                for cm in (self._context_managers.values() or ())
            )
        )

    def _count_waiters(self) -> Dict[PT, int]:
        """Counts the number of waiters for each priority.

        Returns:
            A dictionary mapping each priority to the number of waiters.
        """
        return {
            manager._priority: len(manager.waiters)
            for manager in sorted(self._waiters, key=lambda m: m._priority)
        }

    def _wake_up_next(self) -> None:
        """Wakes up the next waiter in line.

        This method handles the waking of waiters based on priority. It includes an emergency
        procedure to handle potential lost waiters, ensuring that no waiter is left indefinitely
        waiting.
        """
        while self._waiters:
            manager = heapq.heappop(self._waiters)
            if len(manager) == 0:
                # There are no more waiters, get rid of the empty manager
                logger.debug(
                    "manager %s has no more waiters, popping from %s",
                    manager._repr_no_parent_(),
                    self,
                )
                self._context_managers.pop(manager._priority)
                continue
            logger.debug("waking up next for %s", manager._repr_no_parent_())

            woke_up = False
            start_len = len(manager)

            if not manager._waiters:
                logger.debug("not manager._waiters")

            while manager._waiters:
                waiter = manager._waiters.popleft()
                self._potential_lost_waiters.remove(waiter)
                if not waiter.done():
                    waiter.set_result(None)
                    logger.debug("woke up %s", waiter)
                    woke_up = True
                    break

            if not woke_up:
                self._context_managers.pop(manager._priority)
                continue

            end_len = len(manager)

            assert start_len > end_len, f"start {start_len} end {end_len}"

            if end_len:
                # There are still waiters, put the manager back
                heapq.heappush(self._waiters, manager)  # type: ignore [misc]
            else:
                # There are no more waiters, get rid of the empty manager
                self._context_managers.pop(manager._priority)
            return

        # emergency procedure (hopefully temporary):
        while self._potential_lost_waiters:
            waiter = self._potential_lost_waiters.pop(0)
            logger.debug("we found a lost waiter %s", waiter)
            if not waiter.done():
                waiter.set_result(None)
                logger.debug("woke up lost waiter %s", waiter)
                return
        logger.debug("%s has no waiters to wake", self)


class _AbstractPrioritySemaphoreContextManager(Semaphore, Generic[PT]):
    """
    A context manager for priority semaphore waiters.

    This context manager is associated with a specific priority and handles
    the acquisition and release of the semaphore for waiters with that priority.
    """

    _loop: asyncio.AbstractEventLoop
    _waiters: Deque[asyncio.Future]  # type: ignore [assignment]
    __slots__ = "_parent", "_priority"

    @property
    def _priority_name(self) -> str:
        raise NotImplementedError

    def __init__(
        self,
        parent: _AbstractPrioritySemaphore,
        priority: PT,
        name: Optional[str] = None,
    ) -> None:
        """Initializes the context manager for a specific priority.

        Args:
            parent: The parent semaphore.
            priority: The priority associated with this context manager.
            name: An optional name for the context manager, used for debugging.
        """

        self._parent = parent
        """The parent semaphore."""

        self._priority = priority
        """The priority associated with this context manager."""

        super().__init__(0, name=name)

    def __repr__(self) -> str:
        """Returns a string representation of the context manager."""
        return f"<{self.__class__.__name__} parent={self._parent} {self._priority_name}={self._priority} waiters={len(self)}>"

    def _repr_no_parent_(self) -> str:
        """Returns a string representation of the context manager without the parent."""
        return f"<{self.__class__.__name__} parent_name={self._parent.name} {self._priority_name}={self._priority} waiters={len(self)}>"

    def __lt__(self, other) -> bool:
        """Compares this context manager with another based on priority.

        Args:
            other: The other context manager to compare with.

        Returns:
            True if this context manager has a lower priority than the other, False otherwise.
        """
        if type(other) is not type(self):
            raise TypeError(f"{other} is not type {self.__class__.__name__}")
        return self._priority < other._priority

    @cached_property
    def loop(self) -> asyncio.AbstractEventLoop:
        """Gets the event loop associated with this context manager."""
        return self._loop or asyncio.get_event_loop()

    @property
    def waiters(self) -> Deque[asyncio.Future]:
        """Gets the deque of waiters for this context manager."""
        if self._waiters is None:
            self._waiters = deque()
        return self._waiters

    async def acquire(self) -> Literal[True]:
        """Acquires the semaphore for this context manager.

        If the internal counter is larger than zero on entry,
        decrement it by one and return True immediately. If it is
        zero on entry, block, waiting until some other coroutine has
        called release() to make it larger than 0, and then return
        True.
        """
        if self._parent._value <= 0:
            self._ensure_debug_daemon()
        while self._parent._value <= 0:
            fut = self.loop.create_future()
            self.waiters.append(fut)
            self._parent._potential_lost_waiters.append(fut)
            try:
                await fut
            except:
                # See the similar code in Queue.get.
                fut.cancel()
                if self._parent._value > 0 and not fut.cancelled():
                    self._parent._wake_up_next()
                raise
        self._parent._value -= 1
        return True

    def release(self) -> None:
        """Releases the semaphore for this context manager."""
        self._parent.release()


class _PrioritySemaphoreContextManager(
    _AbstractPrioritySemaphoreContextManager[Numeric]
):
    """Context manager for numeric priority semaphores."""

    _priority_name = "priority"


class PrioritySemaphore(_AbstractPrioritySemaphore[Numeric, _PrioritySemaphoreContextManager]):  # type: ignore [type-var]
    """Semaphore that uses numeric priorities for waiters.

    It's similar to a regular Semaphore but requires each waiter to have a priority:

    Examples:
        The primary way to use this semaphore is by specifying a priority.

        >>> priority_semaphore = PrioritySemaphore(10)
        >>> async with priority_semaphore[priority]:
        ...     await do_stuff()

        You can also enter and exit this semaphore without specifying a priority, and it will use the top priority by default:

        >>> priority_semaphore = PrioritySemaphore(10)
        >>> async with priority_semaphore:
        ...     await do_stuff()
    """

    _context_manager_class = _PrioritySemaphoreContextManager
    _top_priority = -1



# mypy: disable-error-code=valid-type
import functools

from a_sync._typing import *
from a_sync.a_sync.config import user_set_default_modifiers, null_modifiers
from a_sync.a_sync.modifiers import cache, limiter, semaphores

valid_modifiers = [
    key
    for key in ModifierKwargs.__annotations__
    if not key.startswith("_") and not key.endswith("_")
]


class ModifierManager(Dict[str, Any]):
    """Manages modifiers for asynchronous and synchronous functions.

    This class is responsible for applying modifiers to functions, such as
    caching, rate limiting, and semaphores for asynchronous functions.
    """

    default: DefaultMode
    cache_type: CacheType
    cache_typed: bool
    ram_cache_maxsize: Optional[int]
    ram_cache_ttl: Optional[int]
    runs_per_minute: Optional[int]
    semaphore: SemaphoreSpec
    # sync modifiers
    executor: Executor
    """This is not applied like a typical modifier. The executor is used to run the sync function in an asynchronous context."""

    __slots__ = ("_modifiers",)

    def __init__(self, modifiers: ModifierKwargs) -> None:
        """Initializes the ModifierManager with the given modifiers.

        Args:
            modifiers: A dictionary of modifiers to be applied.

        Raises:
            ValueError: If an unsupported modifier is provided.
        """
        for key in modifiers.keys():
            if key not in valid_modifiers:
                raise ValueError(f"'{key}' is not a supported modifier.")
        self._modifiers = modifiers

    def __repr__(self) -> str:
        """Returns a string representation of the modifiers."""
        return str(self._modifiers)

    def __getattribute__(self, modifier_key: str) -> Any:
        """Gets the value of a modifier.

        Args:
            modifier_key: The key of the modifier to retrieve.

        Returns:
            The value of the modifier, or the default value if not set.
        """
        if modifier_key not in valid_modifiers:
            return super().__getattribute__(modifier_key)
        return (
            self[modifier_key] if modifier_key in self else user_defaults[modifier_key]
        )

    @property
    def use_limiter(self) -> bool:
        """Determines if a rate limiter should be used."""
        return self.runs_per_minute != nulls.runs_per_minute

    @property
    def use_semaphore(self) -> bool:
        """Determines if a semaphore should be used."""
        return self.semaphore != nulls.semaphore

    @property
    def use_cache(self) -> bool:
        """Determines if caching should be used."""
        return any(
            [
                self.cache_type != nulls.cache_type,
                self.ram_cache_maxsize != nulls.ram_cache_maxsize,
                self.ram_cache_ttl != nulls.ram_cache_ttl,
                self.cache_typed != nulls.cache_typed,
            ]
        )

    def apply_async_modifiers(self, coro_fn: CoroFn[P, T]) -> CoroFn[P, T]:
        """Applies asynchronous modifiers to a coroutine function.

        Args:
            coro_fn: The coroutine function to modify.

        Returns:
            The modified coroutine function.
        """
        # NOTE: THESE STACK IN REVERSE ORDER
        if self.use_limiter:
            coro_fn = limiter.apply_rate_limit(coro_fn, self.runs_per_minute)
        if self.use_semaphore:
            coro_fn = semaphores.apply_semaphore(coro_fn, self.semaphore)
        if self.use_cache:
            coro_fn = cache.apply_async_cache(
                coro_fn,
                cache_type=self.cache_type or "memory",
                cache_typed=self.cache_typed,
                ram_cache_maxsize=self.ram_cache_maxsize,
                ram_cache_ttl=self.ram_cache_ttl,
            )
        return coro_fn

    def apply_sync_modifiers(self, function: SyncFn[P, T]) -> SyncFn[P, T]:
        """Wraps a synchronous function.

        Note:
            There are no sync modifiers at this time, but they will be added here for convenience.

        Args:
            function: The synchronous function to wrap.

        Returns:
            The wrapped synchronous function.
        """

        @functools.wraps(function)
        def sync_modifier_wrap(*args: P.args, **kwargs: P.kwargs) -> T:
            return function(*args, **kwargs)

        return sync_modifier_wrap

    # Dictionary api
    def keys(self) -> KeysView[str]:  # type: ignore [override]
        """Returns the keys of the modifiers."""
        return self._modifiers.keys()

    def values(self) -> ValuesView[Any]:  # type: ignore [override]
        """Returns the values of the modifiers."""
        return self._modifiers.values()

    def items(self) -> ItemsView[str, Any]:  # type: ignore [override]
        """Returns the items of the modifiers."""
        return self._modifiers.items()

    def __contains__(self, key: str) -> bool:  # type: ignore [override]
        """Checks if a key is in the modifiers."""
        return key in self._modifiers

    def __iter__(self) -> Iterator[str]:
        """Returns an iterator over the modifier keys."""
        return self._modifiers.__iter__()

    def __len__(self) -> int:
        """Returns the number of modifiers."""
        return len(self._modifiers)

    def __getitem__(self, modifier_key: str):
        """Gets the value of a modifier by key.

        Args:
            modifier_key: The key of the modifier to retrieve.

        Returns:
            The value of the modifier.
        """
        return self._modifiers[modifier_key]  # type: ignore [literal-required]


nulls = ModifierManager(null_modifiers)
user_defaults = ModifierManager(user_set_default_modifiers)



from a_sync.a_sync._meta import ASyncSingletonMeta
from a_sync.a_sync.base import ASyncGenericBase


class ASyncGenericSingleton(ASyncGenericBase, metaclass=ASyncSingletonMeta):
    """
    A base class for creating singleton-esque ASync classes.

    This class combines the functionality of ASyncGenericBase with a singleton pattern,
    ensuring that only one instance of the class exists per execution mode (sync/async).
    It uses a custom metaclass to manage instance creation and caching.

    Subclasses of ASyncGenericSingleton will have two instances instead of one:
    - one synchronous instance
    - one asynchronous instance

    This allows for proper behavior in both synchronous and asynchronous contexts
    while maintaining the singleton pattern within each context.

    Example:
        class MyAsyncSingleton(ASyncGenericSingleton):
            @a_sync
            def my_method(self):
                # Method implementation

        # These will return the same synchronous instance
        sync_instance1 = MyAsyncSingleton(sync=True)
        sync_instance2 = MyAsyncSingleton(sync=True)

        # These will return the same asynchronous instance
        async_instance1 = MyAsyncSingleton(asynchronous=True)
        async_instance2 = MyAsyncSingleton(asynchronous=True)

        assert sync_instance1 is sync_instance2
        assert async_instance1 is async_instance2
        assert sync_instance1 is not async_instance1
    """



"""
This module provides two specialized async flow management classes, CounterLock and CounterLockCluster.

These primitives manage :class:`asyncio.Task` objects that must wait for an internal counter to reach a specific value.
"""

import asyncio
from collections import defaultdict
from time import time
from typing import DefaultDict, Iterable, Optional

from a_sync.primitives._debug import _DebugDaemonMixin
from a_sync.primitives.locks.event import Event


class CounterLock(_DebugDaemonMixin):
    """
    An async primitive that uses an internal counter to manage task synchronization.

    A coroutine can `await counter.wait_for(3)` and it will wait until the internal counter >= 3.
    If some other task executes `counter.value = 5` or `counter.set(5)`, the first coroutine will proceed as 5 >= 3.

    The internal counter can only be set to a value greater than the current value.
    """

    __slots__ = "is_ready", "_name", "_value", "_events"

    def __init__(self, start_value: int = 0, name: Optional[str] = None):
        """
        Initializes the CounterLock with a starting value and an optional name.

        Args:
            start_value: The initial value of the counter.
            name: An optional name for the counter, used in debug logs.
        """
        self._name = name
        """An optional name for the counter, used in debug logs."""

        self._value = start_value
        """The current value of the counter."""

        self._events: DefaultDict[int, Event] = defaultdict(Event)
        """A defaultdict that maps each awaited value to an :class:`asyncio.Event` that manages the waiters for that value."""

        self.is_ready = lambda v: self._value >= v
        """A lambda function that indicates whether a given value has already been surpassed."""

    async def wait_for(self, value: int) -> bool:
        """
        Waits until the counter reaches or exceeds the specified value.

        Args:
            value: The value to wait for.
        """
        if not self.is_ready(value):
            self._ensure_debug_daemon()
            await self._events[value].wait()
        return True

    def set(self, value: int) -> None:
        """
        Sets the counter to the specified value.

        Args:
            value: The value to set the counter to. Must be strictly greater than the current value.

        Raises:
            ValueError: If the new value is less than or equal to the current value.
        """
        self.value = value

    def __repr__(self) -> str:
        waiters = {v: len(self._events[v]._waiters) for v in sorted(self._events)}
        return f"<CounterLock name={self._name} value={self._value} waiters={waiters}>"

    @property
    def value(self) -> int:
        """
        Gets the current value of the counter.
        """
        return self._value

    @value.setter
    def value(self, value: int) -> None:
        """
        Sets the counter to a new value, waking up any waiters if the value increases beyond the value they are awaiting.

        Args:
            value: The new value of the counter.

        Raises:
            ValueError: If the new value is less than the current value.
        """
        if value > self._value:
            self._value = value
            ready = [
                self._events.pop(key)
                for key in list(self._events.keys())
                if key <= self._value
            ]
            for event in ready:
                event.set()
        elif value < self._value:
            raise ValueError("You cannot decrease the value.")

    async def _debug_daemon(self) -> None:
        """
        Periodically logs debug information about the counter state and waiters.
        """
        start = time()
        while self._events:
            self.logger.debug(
                "%s is still locked after %sm", self, round(time() - start / 60, 2)
            )
            await asyncio.sleep(300)


class CounterLockCluster:
    """
    An asyncio primitive that represents a collection of CounterLock objects.

    `wait_for(i)` will wait until the value of all CounterLock objects is >= i.
    """

    __slots__ = ("locks",)

    def __init__(self, counter_locks: Iterable[CounterLock]) -> None:
        """
        Initializes the CounterLockCluster with a collection of CounterLock objects.

        Args:
            counter_locks: The CounterLock objects to manage.
        """
        self.locks = list(counter_locks)

    async def wait_for(self, value: int) -> bool:
        """
        Waits until the value of all CounterLock objects in the cluster reaches or exceeds the specified value.

        Args:
            value: The value to wait for.
        """
        await asyncio.gather(
            *[counter_lock.wait_for(value) for counter_lock in self.locks]
        )
        return True



"""
This module provides an abstract base class for defining asynchronous and synchronous behavior.

The ASyncABC class uses the ASyncMeta metaclass to automatically wrap its methods
with asynchronous or synchronous behavior based on flags. Subclasses must
implement the abstract methods to define the flag name, flag value, and
default mode for asynchronous or synchronous execution.

Note: It is recommended to use ASyncGenericBase for most use cases. This class
is intended for more custom implementations if necessary.
"""

import abc
import functools
import logging

from a_sync import exceptions
from a_sync._typing import *
from a_sync.a_sync import _flags, _kwargs, modifiers
from a_sync.a_sync._meta import ASyncMeta
from a_sync.exceptions import NoFlagsFound

logger = logging.getLogger(__name__)


class ASyncABC(metaclass=ASyncMeta):
    """Abstract Base Class for defining asynchronous and synchronous behavior.

    This class uses the ASyncMeta metaclass to automatically wrap its methods
    with asynchronous or synchronous behavior based on flags. Subclasses must
    implement the abstract methods to define the flag name, flag value, and
    default mode for asynchronous or synchronous execution.
    """

    ##################################
    # Concrete Methods (overridable) #
    ##################################

    def __a_sync_should_await__(self, kwargs: dict) -> bool:
        """Determines if methods should be called asynchronously.

        This method first checks the provided keyword arguments for flags
        indicating the desired execution mode. If no flags are found, it
        defaults to the instance's asynchronous flag.

        Args:
            kwargs: A dictionary of keyword arguments to check for flags.
        """
        try:
            return self.__a_sync_should_await_from_kwargs__(kwargs)
        except exceptions.NoFlagsFound:
            return self.__a_sync_instance_should_await__

    @functools.cached_property
    def __a_sync_instance_should_await__(self) -> bool:
        """Indicates if the instance should default to asynchronous execution.

        This property can be overridden if dynamic behavior is needed. For
        instance, to allow hot-swapping of instance modes, redefine this as a
        non-cached property.
        """
        return _flags.negate_if_necessary(
            self.__a_sync_flag_name__, self.__a_sync_flag_value__
        )

    def __a_sync_should_await_from_kwargs__(self, kwargs: dict) -> bool:
        """Determines execution mode from keyword arguments.

        This method can be overridden to customize how flags are extracted
        from keyword arguments.

        Args:
            kwargs: A dictionary of keyword arguments to check for flags.

        Raises:
            NoFlagsFound: If no valid flags are found in the keyword arguments.
        """
        if flag := _kwargs.get_flag_name(kwargs):
            return _kwargs.is_sync(flag, kwargs, pop_flag=True)  # type: ignore [arg-type]
        raise NoFlagsFound("kwargs", kwargs.keys())

    @classmethod
    def __a_sync_instance_will_be_sync__(cls, args: tuple, kwargs: dict) -> bool:
        """Determines if a new instance will be synchronous.

        This method checks the constructor's signature against provided
        keyword arguments to determine the execution mode for the new instance.

        Args:
            args: A tuple of positional arguments for the instance.
            kwargs: A dictionary of keyword arguments for the instance.
        """
        logger.debug(
            "checking `%s.%s.__init__` signature against provided kwargs to determine a_sync mode for the new instance",
            cls.__module__,
            cls.__name__,
        )
        if flag := _kwargs.get_flag_name(kwargs):
            sync = _kwargs.is_sync(flag, kwargs)  # type: ignore [arg-type]
            logger.debug(
                "kwargs indicate the new instance created with args %s %s is %ssynchronous",
                args,
                kwargs,
                "a" if sync is False else "",
            )
            return sync
        logger.debug(
            "No valid flags found in kwargs, checking class definition for defined default"
        )
        return cls.__a_sync_default_mode__()  # type: ignore [arg-type]

    ######################################
    # Concrete Methods (non-overridable) #
    ######################################

    @property
    def __a_sync_modifiers__(self: "ASyncABC") -> ModifierKwargs:
        """Retrieves modifiers for the instance.

        This method should not be overridden. It returns the modifiers
        associated with the instance, which are used to customize behavior.
        """
        return modifiers.get_modifiers_from(self)

    ####################
    # Abstract Methods #
    ####################

    @property
    @abc.abstractmethod
    def __a_sync_flag_name__(self) -> str:
        """Abstract property for the flag name.

        Subclasses must implement this property to return the name of the flag
        used to determine execution mode.
        """

    @property
    @abc.abstractmethod
    def __a_sync_flag_value__(self) -> bool:
        """Abstract property for the flag value.

        Subclasses must implement this property to return the value of the flag
        indicating the default execution mode.
        """

    @classmethod
    @abc.abstractmethod  # type: ignore [arg-type, misc]
    def __a_sync_default_mode__(cls) -> bool:  # type: ignore [empty-body]
        """Abstract class method for the default execution mode.

        Subclasses must implement this method to return the default execution
        mode (synchronous or asynchronous) for instances of the class.
        """



"""
This module provides utility functions for handling asynchronous operations
and converting synchronous functions to asynchronous ones.
"""

import asyncio
import functools
from concurrent.futures import Executor

import a_sync.asyncio
from a_sync import exceptions
from a_sync._typing import *


def _await(awaitable: Awaitable[T]) -> T:
    """
    Await an awaitable object in a synchronous context.

    Args:
        awaitable: The awaitable object to be awaited.

    Raises:
        exceptions.SyncModeInAsyncContextError: If the event loop is already running.
    """
    try:
        return a_sync.asyncio.get_event_loop().run_until_complete(awaitable)
    except RuntimeError as e:
        if str(e) == "This event loop is already running":
            raise exceptions.SyncModeInAsyncContextError from None
        raise


def _asyncify(func: SyncFn[P, T], executor: Executor) -> CoroFn[P, T]:  # type: ignore [misc]
    """
    Convert a synchronous function to a coroutine function.

    Args:
        func: The synchronous function to be converted.
        executor: The executor used to run the synchronous function.

    Returns:
        A coroutine function wrapping the input function.

    Raises:
        exceptions.FunctionNotSync: If the input function is a coroutine function or an instance of ASyncFunction.
    """
    from a_sync.a_sync.function import ASyncFunction

    if asyncio.iscoroutinefunction(func) or isinstance(func, ASyncFunction):
        raise exceptions.FunctionNotSync(func)

    @functools.wraps(func)
    async def _asyncify_wrap(*args: P.args, **kwargs: P.kwargs) -> T:
        return await asyncio.futures.wrap_future(
            executor.submit(func, *args, **kwargs),
            loop=a_sync.asyncio.get_event_loop(),
        )

    return _asyncify_wrap



"""
This module includes both new primitives and modified versions of standard asyncio primitives.

The primitives provided in this module are:

- Semaphore
- ThreadsafeSemaphore
- PrioritySemaphore
- CounterLock
- Event
- Queue
- ProcessingQueue
- SmartProcessingQueue

These primitives extend or modify the functionality of standard asyncio primitives to provide additional features or improved performance for specific use cases.
"""

from a_sync.primitives.locks import *
from a_sync.primitives.queue import Queue, ProcessingQueue, SmartProcessingQueue

__all__ = [
    "Semaphore",
    "ThreadsafeSemaphore",
    "PrioritySemaphore",
    "CounterLock",
    "Event",
    "Queue",
    "ProcessingQueue",
    "SmartProcessingQueue",
]



"""
This module provides various semaphore implementations, including a debug-enabled semaphore,
a dummy semaphore that does nothing, and a threadsafe semaphore for use in multi-threaded applications.
"""

import asyncio
import functools
import logging
import sys
from collections import defaultdict
from threading import Thread, current_thread

from a_sync._typing import *
from a_sync.primitives._debug import _DebugDaemonMixin

logger = logging.getLogger(__name__)


class Semaphore(asyncio.Semaphore, _DebugDaemonMixin):
    """
    A semaphore with additional debugging capabilities.

    This semaphore includes debug logging and can be used to decorate coroutine functions.
    It allows rewriting the pattern of acquiring a semaphore within a coroutine using a decorator.

    So you can write this pattern:

    ```
    semaphore = Semaphore(5)

    async def limited():
        async with semaphore:
            return 1

    ```

    like this:

    ```
    semaphore = Semaphore(5)

    @semaphore
    async def limited():
        return 1
    ```
    """

    if sys.version_info >= (3, 10):
        __slots__ = "name", "_value", "_waiters", "_decorated"
    else:
        __slots__ = "name", "_value", "_waiters", "_loop", "_decorated"

    def __init__(self, value: int, name=None, **kwargs) -> None:
        """
        Initialize the semaphore with a given value and optional name for debugging.

        Args:
            value: The initial value for the semaphore.
            name (optional): An optional name used only to provide useful context in debug logs.
        """
        super().__init__(value, **kwargs)
        self.name = name or self.__origin__ if hasattr(self, "__origin__") else None
        self._decorated: Set[str] = set()

    def __call__(self, fn: CoroFn[P, T]) -> CoroFn[P, T]:
        """
        Decorator method to wrap coroutine functions with the semaphore.

        This allows rewriting the pattern of acquiring a semaphore within a coroutine using a decorator.

        Example:
            semaphore = Semaphore(5)

            @semaphore
            async def limited():
                return 1
        """
        return self.decorate(fn)  # type: ignore [arg-type, return-value]

    def __repr__(self) -> str:
        representation = f"<{self.__class__.__name__} name={self.name} value={self._value} waiters={len(self)}>"
        if self._decorated:
            representation = f"{representation[:-1]} decorates={self._decorated}"
        return representation

    def __len__(self) -> int:
        return len(self._waiters) if self._waiters else 0

    def decorate(self, fn: CoroFn[P, T]) -> CoroFn[P, T]:
        """
        Wrap a coroutine function to ensure it runs with the semaphore.

        Example:
            semaphore = Semaphore(5)

            @semaphore
            async def limited():
                return 1
        """
        if not asyncio.iscoroutinefunction(fn):
            raise TypeError(f"{fn} must be a coroutine function")

        @functools.wraps(fn)
        async def semaphore_wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            async with self:
                return await fn(*args, **kwargs)

        self._decorated.add(f"{fn.__module__}.{fn.__name__}")
        return semaphore_wrapper

    async def acquire(self) -> Literal[True]:
        if self._value <= 0:
            self._ensure_debug_daemon()
        return await super().acquire()

    async def _debug_daemon(self) -> None:
        """
        Daemon coroutine (runs in a background task) which will emit a debug log every minute while the semaphore has waiters.
        """
        while self._waiters:
            await asyncio.sleep(60)
            self.logger.debug(
                f"{self} has {len(self)} waiters for any of: {self._decorated}"
            )


class DummySemaphore(asyncio.Semaphore):
    """
    A dummy semaphore that implements the standard :class:`asyncio.Semaphore` API but does nothing.

    This class is useful for scenarios where a semaphore interface is required but no actual synchronization is needed.
    """

    __slots__ = "name", "_value"

    def __init__(self, name: Optional[str] = None):
        """
        Initialize the dummy semaphore with an optional name.

        Args:
            name (optional): An optional name for the dummy semaphore.
        """
        self.name = name
        self._value = 0

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} name={self.name}>"

    async def acquire(self) -> Literal[True]:
        return True

    def release(self) -> None: ...

    async def __aenter__(self): ...

    async def __aexit__(self, *args): ...


class ThreadsafeSemaphore(Semaphore):
    """
    A semaphore that works in a multi-threaded environment.

    This semaphore ensures that the program functions correctly even when used with multiple event loops.
    It provides a workaround for edge cases involving multiple threads and event loops.
    """

    __slots__ = "semaphores", "dummy"

    def __init__(self, value: Optional[int], name: Optional[str] = None) -> None:
        """
        Initialize the threadsafe semaphore with a given value and optional name.

        Args:
            value: The initial value for the semaphore, should be an integer.
            name (optional): An optional name for the semaphore.
        """
        assert isinstance(value, int), f"{value} should be an integer."
        super().__init__(value, name=name)
        self.semaphores: DefaultDict[Thread, Semaphore] = defaultdict(lambda: Semaphore(value, name=self.name))  # type: ignore [arg-type]
        self.dummy = DummySemaphore(name=name)

    def __len__(self) -> int:
        return sum(len(sem._waiters) for sem in self.semaphores.values())

    @functools.cached_property
    def use_dummy(self) -> bool:
        """
        Determine whether to use a dummy semaphore.

        Returns:
            True if the semaphore value is None, indicating the use of a dummy semaphore.
        """
        return self._value is None

    @property
    def semaphore(self) -> Semaphore:
        """
        Returns the appropriate semaphore for the current thread.

        NOTE: We can't cache this property because we need to check the current thread every time we access it.
        """
        return self.dummy if self.use_dummy else self.semaphores[current_thread()]

    async def __aenter__(self):
        await self.semaphore.acquire()

    async def __aexit__(self, *args):
        self.semaphore.release()



"""
This module provides classes for implementing dual-functional sync/async methods in Python.

It includes descriptors and bound methods that can be used to create flexible
asynchronous interfaces, allowing methods to be called both synchronously and
asynchronously based on various conditions and configurations.
"""

# mypy: disable-error-code=valid-type
# mypy: disable-error-code=misc
import functools
import logging
import weakref
from inspect import isawaitable

from a_sync._typing import *
from a_sync.a_sync import _helpers, _kwargs
from a_sync.a_sync._descriptor import ASyncDescriptor
from a_sync.a_sync.function import (
    ASyncFunction,
    ASyncFunctionAsyncDefault,
    ASyncFunctionSyncDefault,
)

if TYPE_CHECKING:
    from a_sync import TaskMapping
    from a_sync.a_sync.abstract import ASyncABC

logger = logging.getLogger(__name__)


class ASyncMethodDescriptor(ASyncDescriptor[I, P, T]):
    """
    This class provides the core functionality for creating :class:`ASyncBoundMethod` objects,
    which can be used to define methods that can be called both synchronously and asynchronously.
    """

    __wrapped__: AnyFn[P, T]
    """The unbound function which will be bound to an instance when :meth:`__get__` is called."""

    async def __call__(self, instance: I, *args: P.args, **kwargs: P.kwargs) -> T:
        """
        Asynchronously call the method.

        Args:
            instance: The instance the method is bound to.
            *args: Positional arguments.
            **kwargs: Keyword arguments.

        Returns:
            The result of the method call.
        """
        # NOTE: This is only used by TaskMapping atm  # TODO: use it elsewhere
        logger.debug(
            "awaiting %s for instance: %s args: %s kwargs: %s",
            self,
            instance,
            args,
            kwargs,
        )
        return await self.__get__(instance, None)(*args, **kwargs)

    @overload
    def __get__(self, instance: None, owner: Type[I]) -> Self: ...
    @overload
    def __get__(self, instance: I, owner: Type[I]) -> "ASyncBoundMethod[I, P, T]": ...
    def __get__(
        self, instance: Optional[I], owner: Type[I]
    ) -> Union[Self, "ASyncBoundMethod[I, P, T]"]:
        """
        Get the bound method or the descriptor itself.

        Args:
            instance: The instance to bind the method to, or None.
            owner: The owner class.

        Returns:
            The descriptor or bound method.
        """
        if instance is None:
            return self
        try:
            bound = instance.__dict__[self.field_name]
            # we will set a new one in the finally block
            bound._cache_handle.cancel()
        except KeyError:
            from a_sync.a_sync.abstract import ASyncABC

            if self.default == "sync":
                bound = ASyncBoundMethodSyncDefault(
                    instance, self.__wrapped__, self.__is_async_def__, **self.modifiers
                )
            elif self.default == "async":
                bound = ASyncBoundMethodAsyncDefault(
                    instance, self.__wrapped__, self.__is_async_def__, **self.modifiers
                )
            elif isinstance(instance, ASyncABC):
                try:
                    if instance.__a_sync_instance_should_await__:
                        bound = ASyncBoundMethodSyncDefault(
                            instance,
                            self.__wrapped__,
                            self.__is_async_def__,
                            **self.modifiers,
                        )
                    else:
                        bound = ASyncBoundMethodAsyncDefault(
                            instance,
                            self.__wrapped__,
                            self.__is_async_def__,
                            **self.modifiers,
                        )
                except AttributeError:
                    bound = ASyncBoundMethod(
                        instance,
                        self.__wrapped__,
                        self.__is_async_def__,
                        **self.modifiers,
                    )
            else:
                bound = ASyncBoundMethod(
                    instance, self.__wrapped__, self.__is_async_def__, **self.modifiers
                )
            instance.__dict__[self.field_name] = bound
            logger.debug("new bound method: %s", bound)
        # Handler for popping unused bound methods from bound method cache
        bound._cache_handle = self._get_cache_handle(instance)
        return bound

    def __set__(self, instance, value):
        """
        Prevent setting the descriptor.

        Args:
            instance: The instance.
            value: The value to set.

        Raises:
            :class:`RuntimeError`: Always raised to prevent setting.
        """
        raise RuntimeError(
            f"cannot set {self.field_name}, {self} is what you get. sorry."
        )

    def __delete__(self, instance):
        """
        Prevent deleting the descriptor.

        Args:
            instance: The instance.

        Raises:
            :class:`RuntimeError`: Always raised to prevent deletion.
        """
        raise RuntimeError(
            f"cannot delete {self.field_name}, you're stuck with {self} forever. sorry."
        )

    @functools.cached_property
    def __is_async_def__(self) -> bool:
        """
        Check if the wrapped function is a coroutine function.

        Returns:
            True if the wrapped function is a coroutine function, False otherwise.
        """
        return asyncio.iscoroutinefunction(self.__wrapped__)

    def _get_cache_handle(self, instance: I) -> asyncio.TimerHandle:
        """
        Get a cache handle for the instance.

        Args:
            instance: The instance to create a cache handle for.

        Returns:
            A timer handle for cache management.
        """
        # NOTE: use `instance.__dict__.pop` instead of `delattr` so we don't create a strong ref to `instance`
        return asyncio.get_event_loop().call_later(
            300, instance.__dict__.pop, self.field_name
        )


@final
class ASyncMethodDescriptorSyncDefault(ASyncMethodDescriptor[I, P, T]):
    """
    A descriptor for :class:`ASyncBoundMethodSyncDefault` objects.

    This class extends ASyncMethodDescriptor to provide a synchronous
    default behavior for method calls.
    """

    default = "sync"
    """The default mode for this bound method. Always set to "sync"."""

    any: ASyncFunctionSyncDefault[Concatenate[AnyIterable[I], P], bool]
    """Synchronous default version of the :meth:`~ASyncMethodDescriptor.any` method."""

    all: ASyncFunctionSyncDefault[Concatenate[AnyIterable[I], P], bool]
    """Synchronous default version of the :meth:`~ASyncMethodDescriptor.all` method."""

    min: ASyncFunctionSyncDefault[Concatenate[AnyIterable[I], P], T]
    """Synchronous default version of the :meth:`~ASyncMethodDescriptor.min` method."""

    max: ASyncFunctionSyncDefault[Concatenate[AnyIterable[I], P], T]
    """Synchronous default version of the :meth:`~ASyncMethodDescriptor.max` method."""

    sum: ASyncFunctionSyncDefault[Concatenate[AnyIterable[I], P], T]
    """Synchronous default version of the :meth:`~ASyncMethodDescriptor.sum` method."""

    @overload
    def __get__(
        self, instance: None, owner: Type[I] = None
    ) -> "ASyncMethodDescriptorSyncDefault[I, P, T]": ...
    @overload
    def __get__(
        self, instance: I, owner: Type[I] = None
    ) -> "ASyncBoundMethodSyncDefault[I, P, T]": ...
    def __get__(
        self, instance: Optional[I], owner: Type[I] = None
    ) -> (
        "Union[ASyncMethodDescriptorSyncDefault, ASyncBoundMethodSyncDefault[I, P, T]]"
    ):
        """
        Get the bound method or the descriptor itself.

        Args:
            instance: The instance to bind the method to, or None.
            owner: The owner class.

        Returns:
            The descriptor or bound method with synchronous default.
        """
        if instance is None:
            return self
        try:
            bound = instance.__dict__[self.field_name]
            # we will set a new one in the finally block
            bound._cache_handle.cancel()
        except KeyError:
            bound = ASyncBoundMethodSyncDefault(
                instance, self.__wrapped__, self.__is_async_def__, **self.modifiers
            )
            instance.__dict__[self.field_name] = bound
            logger.debug("new bound method: %s", bound)
        # Handler for popping unused bound methods from bound method cache
        bound._cache_handle = self._get_cache_handle(instance)
        return bound


@final
class ASyncMethodDescriptorAsyncDefault(ASyncMethodDescriptor[I, P, T]):
    """
    A descriptor for asynchronous methods with an asynchronous default.

    This class extends ASyncMethodDescriptor to provide an asynchronous default
    behavior for method calls.
    """

    default = "async"
    """The default mode for this bound method. Always set to "async"."""

    any: ASyncFunctionAsyncDefault[Concatenate[AnyIterable[I], P], bool]
    """Asynchronous default version of the :meth:`~ASyncMethodDescriptor.any` method."""

    all: ASyncFunctionAsyncDefault[Concatenate[AnyIterable[I], P], bool]
    """Asynchronous default version of the :meth:`~ASyncMethodDescriptor.all` method."""

    min: ASyncFunctionAsyncDefault[Concatenate[AnyIterable[I], P], T]
    """Asynchronous default version of the :meth:`~ASyncMethodDescriptor.min` method."""

    max: ASyncFunctionAsyncDefault[Concatenate[AnyIterable[I], P], T]
    """Asynchronous default version of the :meth:`~ASyncMethodDescriptor.max` method."""

    sum: ASyncFunctionAsyncDefault[Concatenate[AnyIterable[I], P], T]
    """Asynchronous default version of the :meth:`~ASyncMethodDescriptor.sum` method."""

    @overload
    def __get__(
        self, instance: None, owner: Type[I]
    ) -> "ASyncMethodDescriptorAsyncDefault[I, P, T]": ...
    @overload
    def __get__(
        self, instance: I, owner: Type[I]
    ) -> "ASyncBoundMethodAsyncDefault[I, P, T]": ...
    def __get__(
        self, instance: Optional[I], owner: Type[I]
    ) -> "Union[ASyncMethodDescriptorAsyncDefault, ASyncBoundMethodAsyncDefault[I, P, T]]":
        """
        Get the bound method or the descriptor itself.

        Args:
            instance: The instance to bind the method to, or None.
            owner: The owner class.

        Returns:
            The descriptor or bound method with asynchronous default.
        """
        if instance is None:
            return self
        try:
            bound = instance.__dict__[self.field_name]
            # we will set a new one in the finally block
            bound._cache_handle.cancel()
        except KeyError:
            bound = ASyncBoundMethodAsyncDefault(
                instance, self.__wrapped__, self.__is_async_def__, **self.modifiers
            )
            instance.__dict__[self.field_name] = bound
            logger.debug("new bound method: %s", bound)
        # Handler for popping unused bound methods from bound method cache
        bound._cache_handle = self._get_cache_handle(instance)
        return bound


class ASyncBoundMethod(ASyncFunction[P, T], Generic[I, P, T]):
    """
    A bound method that can be called both synchronously and asynchronously.

    This class represents a method bound to an instance, which can be called
    either synchronously or asynchronously based on various conditions.
    """

    # NOTE: this is created by the Descriptor

    _cache_handle: asyncio.TimerHandle
    "An asyncio handle used to pop the bound method from `instance.__dict__` 5 minutes after its last use."

    __weakself__: "weakref.ref[I]"
    "A weak reference to the instance the function is bound to."

    __wrapped__: AnyFn[Concatenate[I, P], T]
    """The original unbound method that was wrapped."""

    __slots__ = "_is_async_def", "__weakself__"

    def __init__(
        self,
        instance: I,
        unbound: AnyFn[Concatenate[I, P], T],
        async_def: bool,
        **modifiers: Unpack[ModifierKwargs],
    ) -> None:
        """
        Initialize the bound method.

        Args:
            instance: The instance to bind the method to.
            unbound: The unbound function.
            async_def: Whether the original function is an async def.
            **modifiers: Additional modifiers for the function.
        """
        self.__weakself__ = weakref.ref(instance, self.__cancel_cache_handle)
        # First we unwrap the coro_fn and rewrap it so overriding flag kwargs are handled automagically.
        if isinstance(unbound, ASyncFunction):
            modifiers.update(unbound.modifiers)
            unbound = unbound.__wrapped__
        ASyncFunction.__init__(self, unbound, **modifiers)
        self._is_async_def = async_def
        "True if `self.__wrapped__` is a coroutine function, False otherwise."
        functools.update_wrapper(self, unbound)

    def __repr__(self) -> str:
        """
        Return a string representation of the bound method.

        Returns:
            A string representation of the bound method.
        """
        try:
            instance_type = type(self.__self__)
            return f"<{self.__class__.__name__} for function {instance_type.__module__}.{instance_type.__name__}.{self.__name__} bound to {self.__self__}>"
        except ReferenceError:
            return f"<{self.__class__.__name__} for function COLLECTED.COLLECTED.{self.__name__} bound to {self.__weakself__}>"

    @overload
    def __call__(self, *args: P.args, sync: Literal[True], **kwargs: P.kwargs) -> T: ...
    @overload
    def __call__(
        self, *args: P.args, sync: Literal[False], **kwargs: P.kwargs
    ) -> Coroutine[Any, Any, T]: ...
    @overload
    def __call__(
        self, *args: P.args, asynchronous: Literal[False], **kwargs: P.kwargs
    ) -> T: ...
    @overload
    def __call__(
        self, *args: P.args, asynchronous: Literal[True], **kwargs: P.kwargs
    ) -> Coroutine[Any, Any, T]: ...
    @overload
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> MaybeCoro[T]: ...
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> MaybeCoro[T]:
        """
        Call the bound method.

        This method handles both synchronous and asynchronous calls based on
        the provided flags and the method's configuration.

        Args:
            *args: Positional arguments.
            **kwargs: Keyword arguments.

        Returns:
            The result of the method call, which may be a coroutine.
        """
        logger.debug("calling %s with args: %s kwargs: %s", self, args, kwargs)
        # This could either be a coroutine or a return value from an awaited coroutine,
        #   depending on if an overriding flag kwarg was passed into the function call.
        retval = coro = ASyncFunction.__call__(self, self.__self__, *args, **kwargs)
        if not isawaitable(retval):
            # The coroutine was already awaited due to the use of an overriding flag kwarg.
            # We can return the value.
            pass
        elif self._should_await(kwargs):
            # The awaitable was not awaited, so now we need to check the flag as defined on 'self' and await if appropriate.
            logger.debug(
                "awaiting %s for %s args: %s kwargs: %s", coro, self, args, kwargs
            )
            retval = _helpers._await(coro)
        logger.debug(
            "returning %s for %s args: %s kwargs: %s", retval, self, args, kwargs
        )
        return retval  # type: ignore [call-overload, return-value]

    @property
    def __self__(self) -> I:
        """
        Get the instance the method is bound to.

        Returns:
            The instance the method is bound to.

        Raises:
            :class:`ReferenceError`: If the instance has been garbage collected.
        """
        instance = self.__weakself__()
        if instance is not None:
            return instance
        raise ReferenceError(self)

    @functools.cached_property
    def __bound_to_a_sync_instance__(self) -> bool:
        """
        Check if the method is bound to an ASyncABC instance.

        Returns:
            True if bound to an ASyncABC instance, False otherwise.
        """
        from a_sync.a_sync.abstract import ASyncABC

        return isinstance(self.__self__, ASyncABC)

    def map(
        self,
        *iterables: AnyIterable[I],
        concurrency: Optional[int] = None,
        task_name: str = "",
        **kwargs: P.kwargs,
    ) -> "TaskMapping[I, T]":
        """
        Create a TaskMapping for this method.

        Args:
            *iterables: Iterables to map over.
            concurrency: Optional concurrency limit.
            task_name: Optional name for the task.
            **kwargs: Additional keyword arguments.

        Returns:
            A TaskMapping instance for this method.
        """
        from a_sync import TaskMapping

        return TaskMapping(
            self, *iterables, concurrency=concurrency, name=task_name, **kwargs
        )

    async def any(
        self,
        *iterables: AnyIterable[I],
        concurrency: Optional[int] = None,
        task_name: str = "",
        **kwargs: P.kwargs,
    ) -> bool:
        """
        Check if any of the results are truthy.

        Args:
            *iterables: Iterables to map over.
            concurrency: Optional concurrency limit.
            task_name: Optional name for the task.
            **kwargs: Additional keyword arguments.

        Returns:
            True if any result is truthy, False otherwise.
        """
        return await self.map(
            *iterables, concurrency=concurrency, task_name=task_name, **kwargs
        ).any(pop=True, sync=False)

    async def all(
        self,
        *iterables: AnyIterable[I],
        concurrency: Optional[int] = None,
        task_name: str = "",
        **kwargs: P.kwargs,
    ) -> bool:
        """
        Check if all of the results are truthy.

        Args:
            *iterables: Iterables to map over.
            concurrency: Optional concurrency limit.
            task_name: Optional name for the task.
            **kwargs: Additional keyword arguments.

        Returns:
            True if all results are truthy, False otherwise.
        """
        return await self.map(
            *iterables, concurrency=concurrency, task_name=task_name, **kwargs
        ).all(pop=True, sync=False)

    async def min(
        self,
        *iterables: AnyIterable[I],
        concurrency: Optional[int] = None,
        task_name: str = "",
        **kwargs: P.kwargs,
    ) -> T:
        """
        Find the minimum result.

        Args:
            *iterables: Iterables to map over.
            concurrency: Optional concurrency limit.
            task_name: Optional name for the task.
            **kwargs: Additional keyword arguments.

        Returns:
            The minimum result.
        """
        return await self.map(
            *iterables, concurrency=concurrency, task_name=task_name, **kwargs
        ).min(pop=True, sync=False)

    async def max(
        self,
        *iterables: AnyIterable[I],
        concurrency: Optional[int] = None,
        task_name: str = "",
        **kwargs: P.kwargs,
    ) -> T:
        """
        Find the maximum result.

        Args:
            *iterables: Iterables to map over.
            concurrency: Optional concurrency limit.
            task_name: Optional name for the task.
            **kwargs: Additional keyword arguments.

        Returns:
            The maximum result.
        """
        return await self.map(
            *iterables, concurrency=concurrency, task_name=task_name, **kwargs
        ).max(pop=True, sync=False)

    async def sum(
        self,
        *iterables: AnyIterable[I],
        concurrency: Optional[int] = None,
        task_name: str = "",
        **kwargs: P.kwargs,
    ) -> T:
        """
        Calculate the sum of the results.

        Args:
            *iterables: Iterables to map over.
            concurrency: Optional concurrency limit.
            task_name: Optional name for the task.
            **kwargs: Additional keyword arguments.

        Returns:
            The sum of the results.
        """
        return await self.map(
            *iterables, concurrency=concurrency, task_name=task_name, **kwargs
        ).sum(pop=True, sync=False)

    def _should_await(self, kwargs: dict) -> bool:
        """
        Determine if the method should be awaited.

        Args:
            kwargs: Keyword arguments passed to the method.

        Returns:
            True if the method should be awaited, False otherwise.
        """
        if flag := _kwargs.get_flag_name(kwargs):
            return _kwargs.is_sync(flag, kwargs, pop_flag=True)  # type: ignore [arg-type]
        elif self.default:
            return self.default == "sync"
        elif self.__bound_to_a_sync_instance__:
            self.__self__: "ASyncABC"
            return self.__self__.__a_sync_should_await__(kwargs)
        return self._is_async_def

    def __cancel_cache_handle(self, instance: I) -> None:
        """
        Cancel the cache handle.

        Args:
            instance: The instance associated with the cache handle.
        """
        cache_handle: asyncio.TimerHandle = self._cache_handle
        cache_handle.cancel()


class ASyncBoundMethodSyncDefault(ASyncBoundMethod[I, P, T]):
    """
    A bound method with synchronous default behavior.
    """

    def __get__(
        self, instance: Optional[I], owner: Type[I]
    ) -> ASyncFunctionSyncDefault[P, T]:
        """
        Get the bound method or descriptor.

        Args:
            instance: The instance to bind the method to, or None.
            owner: The owner class.

        Returns:
            The bound method with synchronous default behavior.
        """
        return ASyncBoundMethod.__get__(self, instance, owner)

    @overload
    def __call__(self, *args: P.args, sync: Literal[True], **kwargs: P.kwargs) -> T: ...
    @overload
    def __call__(
        self, *args: P.args, sync: Literal[False], **kwargs: P.kwargs
    ) -> Coroutine[Any, Any, T]: ...
    @overload
    def __call__(
        self, *args: P.args, asynchronous: Literal[False], **kwargs: P.kwargs
    ) -> T: ...
    @overload
    def __call__(
        self, *args: P.args, asynchronous: Literal[True], **kwargs: P.kwargs
    ) -> Coroutine[Any, Any, T]: ...
    @overload
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T: ...
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        """
        Call the bound method with synchronous default behavior.

        Args:
            *args: Positional arguments.
            **kwargs: Keyword arguments.

        Returns:
            The result of the method call.
        """
        return ASyncBoundMethod.__call__(self, *args, **kwargs)


class ASyncBoundMethodAsyncDefault(ASyncBoundMethod[I, P, T]):
    """
    A bound method with asynchronous default behavior.
    """

    def __get__(self, instance: I, owner: Type[I]) -> ASyncFunctionAsyncDefault[P, T]:
        """
        Get the bound method or descriptor.

        Args:
            instance: The instance to bind the method to.
            owner: The owner class.

        Returns:
            The bound method with asynchronous default behavior.
        """
        return ASyncBoundMethod.__get__(self, instance, owner)

    @overload
    def __call__(self, *args: P.args, sync: Literal[True], **kwargs: P.kwargs) -> T: ...
    @overload
    def __call__(
        self, *args: P.args, sync: Literal[False], **kwargs: P.kwargs
    ) -> Coroutine[Any, Any, T]: ...
    @overload
    def __call__(
        self, *args: P.args, asynchronous: Literal[False], **kwargs: P.kwargs
    ) -> T: ...
    @overload
    def __call__(
        self, *args: P.args, asynchronous: Literal[True], **kwargs: P.kwargs
    ) -> Coroutine[Any, Any, T]: ...
    @overload
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> Coroutine[Any, Any, T]: ...
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> Coroutine[Any, Any, T]:
        """
        Call the bound method with asynchronous default behavior.

        Args:
            *args: Positional arguments.
            **kwargs: Keyword arguments.

        Returns:
            A coroutine representing the asynchronous method call.
        """
        return ASyncBoundMethod.__call__(self, *args, **kwargs)



# mypy: disable-error-code=valid-type
# mypy: disable-error-code=misc
import asyncio

from aiolimiter import AsyncLimiter

from a_sync import aliases, exceptions
from a_sync._typing import *


LimiterSpec = Union[int, AsyncLimiter]


@overload
def apply_rate_limit(
    runs_per_minute: int,
) -> AsyncDecorator[P, T]:
    """Decorator to apply a rate limit to an asynchronous function.

    Args:
        runs_per_minute: The number of allowed executions per minute.
    """


@overload
def apply_rate_limit(
    coro_fn: CoroFn[P, T],
    runs_per_minute: LimiterSpec,
) -> CoroFn[P, T]:
    """Decorator to apply a rate limit to an asynchronous function.

    Args:
        coro_fn: The coroutine function to be rate-limited.
        runs_per_minute: The number of allowed executions per minute or an AsyncLimiter instance.
    """


def apply_rate_limit(
    coro_fn: Optional[Union[CoroFn[P, T], int]] = None,
    runs_per_minute: Optional[LimiterSpec] = None,
) -> AsyncDecoratorOrCoroFn[P, T]:
    """Applies a rate limit to an asynchronous function.

    This function can be used as a decorator to limit the number of times
    an asynchronous function can be called per minute. It can be configured
    with either an integer specifying the number of runs per minute or an
    AsyncLimiter instance.

    Args:
        coro_fn: The coroutine function to be rate-limited. If an integer is provided, it is treated as runs per minute, and runs_per_minute should be None.
        runs_per_minute: The number of allowed executions per minute or an AsyncLimiter instance. If coro_fn is an integer, this should be None.

    Raises:
        TypeError: If 'runs_per_minute' is neither an integer nor an AsyncLimiter when 'coro_fn' is None.
        exceptions.FunctionNotAsync: If 'coro_fn' is not an asynchronous function.
    """
    # Parse Inputs
    if isinstance(coro_fn, (int, AsyncLimiter)):
        assert runs_per_minute is None
        runs_per_minute = coro_fn
        coro_fn = None

    elif coro_fn is None:
        if runs_per_minute is not None and not isinstance(
            runs_per_minute, (int, AsyncLimiter)
        ):
            raise TypeError("'runs_per_minute' must be an integer.", runs_per_minute)

    elif not asyncio.iscoroutinefunction(coro_fn):
        raise exceptions.FunctionNotAsync(coro_fn)

    def rate_limit_decorator(coro_fn: CoroFn[P, T]) -> CoroFn[P, T]:
        limiter = (
            runs_per_minute
            if isinstance(runs_per_minute, AsyncLimiter)
            else AsyncLimiter(runs_per_minute) if runs_per_minute else aliases.dummy
        )

        async def rate_limit_wrap(*args: P.args, **kwargs: P.kwargs) -> T:
            async with limiter:  # type: ignore [attr-defined]
                return await coro_fn(*args, **kwargs)

        return rate_limit_wrap

    return rate_limit_decorator if coro_fn is None else rate_limit_decorator(coro_fn)



"""
This module provides an enhanced version of :func:`asyncio.gather`.
"""

import asyncio
from typing import Any, Awaitable, Dict, List, Mapping, TypeVar, Union, overload

try:
    from tqdm.asyncio import tqdm_asyncio
except ImportError as e:

    class tqdm_asyncio:  # type: ignore [no-redef]
        async def gather(*args, **kwargs):
            raise ImportError(
                "You must have tqdm installed in order to use this feature"
            )


from a_sync._typing import *
from a_sync.asyncio.as_completed import as_completed_mapping, _exc_wrap


Excluder = Callable[[T], bool]


@overload
async def gather(
    *awaitables: Mapping[K, Awaitable[V]],
    return_exceptions: bool = False,
    exclude_if: Optional[Excluder[V]] = None,
    tqdm: bool = False,
    **tqdm_kwargs: Any,
) -> Dict[K, V]: ...
@overload
async def gather(
    *awaitables: Awaitable[T],
    return_exceptions: bool = False,
    exclude_if: Optional[Excluder[T]] = None,
    tqdm: bool = False,
    **tqdm_kwargs: Any,
) -> List[T]: ...
async def gather(
    *awaitables: Union[Awaitable[T], Mapping[K, Awaitable[V]]],
    return_exceptions: bool = False,
    exclude_if: Optional[Excluder[T]] = None,
    tqdm: bool = False,
    **tqdm_kwargs: Any,
) -> Union[List[T], Dict[K, V]]:
    """
    Concurrently awaits a list of awaitable objects or a mapping of awaitables and returns the results.

    This function extends Python's asyncio.gather, providing additional features for handling either individual awaitable objects or a mapping of awaitables.

    Differences from asyncio.gather:
    - Uses type hints for use with static type checkers.
    - Supports gathering either individual awaitables or a k:v mapping of awaitables.
    - Provides progress reporting using tqdm if 'tqdm' is set to True.
    - Allows exclusion of results based on a condition using the 'exclude_if' parameter.

    Args:
        *awaitables: The awaitables to await concurrently. It can be a list of individual awaitables or a mapping of awaitables.
        return_exceptions: If True, exceptions are returned as results instead of raising them. Defaults to False.
        exclude_if: A callable that takes a result and returns True if the result should be excluded from the final output. Defaults to None.
        tqdm: If True, enables progress reporting using tqdm. Defaults to False.
        **tqdm_kwargs: Additional keyword arguments for tqdm if progress reporting is enabled.

    Examples:
        Awaiting individual awaitables:

        - Results will be a list containing the result of each awaitable in sequential order.

        >>> results = await gather(thing1(), thing2())
        >>> results
        ['result', 123]

        Awaiting a mapping of awaitables:

        - Results will be a dictionary with 'key1' mapped to the result of thing1() and 'key2' mapped to the result of thing2.

        >>> mapping = {'key1': thing1(), 'key2': thing2()}
        >>> results = await gather(mapping)
        >>> results
        {'key1': 'result', 'key2': 123}
    """
    is_mapping = _is_mapping(awaitables)
    results = await (
        gather_mapping(
            awaitables[0],
            return_exceptions=return_exceptions,
            exclude_if=exclude_if,
            tqdm=tqdm,
            **tqdm_kwargs,
        )
        if is_mapping
        else (
            tqdm_asyncio.gather(
                *(
                    (_exc_wrap(a) for a in awaitables)
                    if return_exceptions
                    else awaitables
                ),
                **tqdm_kwargs,
            )
            if tqdm
            else asyncio.gather(*awaitables, return_exceptions=return_exceptions)
        )  # type: ignore [arg-type]
    )
    if exclude_if and not is_mapping:
        results = [r for r in results if not exclude_if(r)]
    return results


async def gather_mapping(
    mapping: Mapping[K, Awaitable[V]],
    return_exceptions: bool = False,
    exclude_if: Optional[Excluder[V]] = None,
    tqdm: bool = False,
    **tqdm_kwargs: Any,
) -> Dict[K, V]:
    """
    Concurrently awaits a mapping of awaitable objects and returns a dictionary of results.

    This function is designed to await a mapping of awaitable objects, where each key-value pair represents a unique awaitable. It enables concurrent execution and gathers results into a dictionary.

    Args:
        mapping: A dictionary-like object where keys are of type K and values are awaitable objects of type V.
        return_exceptions: If True, exceptions are returned as results instead of raising them. Defaults to False.
        exclude_if: A callable that takes a result and returns True if the result should be excluded from the final output. Defaults to None.
        tqdm: If True, enables progress reporting using tqdm. Defaults to False.
        **tqdm_kwargs: Additional keyword arguments for tqdm if progress reporting is enabled.

    Example:
        The 'results' dictionary will contain the awaited results, where keys match the keys in the 'mapping' and values contain the results of the corresponding awaitables.

        >>> mapping = {'task1': async_function1(), 'task2': async_function2(), 'task3': async_function3()}
        >>> results = await gather_mapping(mapping)
        >>> results
        {'task1': "result", 'task2': 123, 'task3': None}
    """
    results = {
        k: v
        async for k, v in as_completed_mapping(
            mapping,
            return_exceptions=return_exceptions,
            aiter=True,
            tqdm=tqdm,
            **tqdm_kwargs,
        )
        if exclude_if is None or not exclude_if(v)
    }
    # return data in same order as input mapping
    return {k: results[k] for k in mapping}


_is_mapping = lambda awaitables: len(awaitables) == 1 and isinstance(
    awaitables[0], Mapping
)

__all__ = ["gather", "gather_mapping"]
